package org.yaoqiang.bpmn.editor.io.bpmn;

import java.awt.Point;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.yaoqiang.bpmn.editor.util.BPMNEditorUtils;
import org.yaoqiang.bpmn.editor.view.BPMNGraph;
import org.yaoqiang.bpmn.model.BPMNModelUtils;
import org.yaoqiang.bpmn.model.elements.XMLCollection;
import org.yaoqiang.bpmn.model.elements.XMLComplexElement;
import org.yaoqiang.bpmn.model.elements.XMLElement;
import org.yaoqiang.bpmn.model.elements.activities.Activity;
import org.yaoqiang.bpmn.model.elements.activities.AdHocSubProcess;
import org.yaoqiang.bpmn.model.elements.activities.BusinessRuleTask;
import org.yaoqiang.bpmn.model.elements.activities.CallActivity;
import org.yaoqiang.bpmn.model.elements.activities.GlobalBusinessRuleTask;
import org.yaoqiang.bpmn.model.elements.activities.GlobalScriptTask;
import org.yaoqiang.bpmn.model.elements.activities.LoopCharacteristics;
import org.yaoqiang.bpmn.model.elements.activities.MultiInstanceLoopCharacteristics;
import org.yaoqiang.bpmn.model.elements.activities.ReceiveTask;
import org.yaoqiang.bpmn.model.elements.activities.ScriptTask;
import org.yaoqiang.bpmn.model.elements.activities.SendTask;
import org.yaoqiang.bpmn.model.elements.activities.ServiceTask;
import org.yaoqiang.bpmn.model.elements.activities.SubProcess;
import org.yaoqiang.bpmn.model.elements.activities.Task;
import org.yaoqiang.bpmn.model.elements.activities.Transaction;
import org.yaoqiang.bpmn.model.elements.artifacts.Association;
import org.yaoqiang.bpmn.model.elements.artifacts.Category;
import org.yaoqiang.bpmn.model.elements.artifacts.CategoryValue;
import org.yaoqiang.bpmn.model.elements.artifacts.Group;
import org.yaoqiang.bpmn.model.elements.artifacts.TextAnnotation;
import org.yaoqiang.bpmn.model.elements.bpmndi.BPMNDiagram;
import org.yaoqiang.bpmn.model.elements.bpmndi.BPMNEdge;
import org.yaoqiang.bpmn.model.elements.bpmndi.BPMNShape;
import org.yaoqiang.bpmn.model.elements.bpmndi.Bounds;
import org.yaoqiang.bpmn.model.elements.bpmndi.Waypoint;
import org.yaoqiang.bpmn.model.elements.choreography.Choreography;
import org.yaoqiang.bpmn.model.elements.choreography.GlobalChoreographyTask;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.CallChoreography;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.ChoreographyActivity;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.ChoreographyTask;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.SubChoreography;
import org.yaoqiang.bpmn.model.elements.collaboration.Collaboration;
import org.yaoqiang.bpmn.model.elements.collaboration.MessageFlow;
import org.yaoqiang.bpmn.model.elements.collaboration.Participant;
import org.yaoqiang.bpmn.model.elements.conversations.CallConversation;
import org.yaoqiang.bpmn.model.elements.conversations.Conversation;
import org.yaoqiang.bpmn.model.elements.conversations.ConversationLink;
import org.yaoqiang.bpmn.model.elements.conversations.ConversationNode;
import org.yaoqiang.bpmn.model.elements.conversations.GlobalConversation;
import org.yaoqiang.bpmn.model.elements.conversations.SubConversation;
import org.yaoqiang.bpmn.model.elements.core.common.BPMNError;
import org.yaoqiang.bpmn.model.elements.core.common.CallableElement;
import org.yaoqiang.bpmn.model.elements.core.common.FlowElement;
import org.yaoqiang.bpmn.model.elements.core.common.FlowNode;
import org.yaoqiang.bpmn.model.elements.core.common.Message;
import org.yaoqiang.bpmn.model.elements.core.common.SequenceFlow;
import org.yaoqiang.bpmn.model.elements.core.foundation.RootElements;
import org.yaoqiang.bpmn.model.elements.core.infrastructure.Definitions;
import org.yaoqiang.bpmn.model.elements.core.service.Interface;
import org.yaoqiang.bpmn.model.elements.core.service.Operation;
import org.yaoqiang.bpmn.model.elements.data.DataAssociation;
import org.yaoqiang.bpmn.model.elements.data.DataInput;
import org.yaoqiang.bpmn.model.elements.data.DataInputAssociation;
import org.yaoqiang.bpmn.model.elements.data.DataObject;
import org.yaoqiang.bpmn.model.elements.data.DataObjectReference;
import org.yaoqiang.bpmn.model.elements.data.DataOutput;
import org.yaoqiang.bpmn.model.elements.data.DataOutputAssociation;
import org.yaoqiang.bpmn.model.elements.data.DataStore;
import org.yaoqiang.bpmn.model.elements.data.DataStoreReference;
import org.yaoqiang.bpmn.model.elements.data.ItemAwareElement;
import org.yaoqiang.bpmn.model.elements.events.BoundaryEvent;
import org.yaoqiang.bpmn.model.elements.events.CancelEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.CatchEvent;
import org.yaoqiang.bpmn.model.elements.events.CompensateEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.ConditionalEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.EndEvent;
import org.yaoqiang.bpmn.model.elements.events.ErrorEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.EscalationEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.Event;
import org.yaoqiang.bpmn.model.elements.events.IntermediateCatchEvent;
import org.yaoqiang.bpmn.model.elements.events.IntermediateThrowEvent;
import org.yaoqiang.bpmn.model.elements.events.LinkEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.MessageEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.SignalEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.StartEvent;
import org.yaoqiang.bpmn.model.elements.events.TerminateEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.ThrowEvent;
import org.yaoqiang.bpmn.model.elements.events.TimerEventDefinition;
import org.yaoqiang.bpmn.model.elements.gateways.ComplexGateway;
import org.yaoqiang.bpmn.model.elements.gateways.EventBasedGateway;
import org.yaoqiang.bpmn.model.elements.gateways.ExclusiveGateway;
import org.yaoqiang.bpmn.model.elements.gateways.Gateway;
import org.yaoqiang.bpmn.model.elements.gateways.InclusiveGateway;
import org.yaoqiang.bpmn.model.elements.gateways.ParallelGateway;
import org.yaoqiang.bpmn.model.elements.humaninteraction.GlobalManualTask;
import org.yaoqiang.bpmn.model.elements.humaninteraction.GlobalUserTask;
import org.yaoqiang.bpmn.model.elements.humaninteraction.ManualTask;
import org.yaoqiang.bpmn.model.elements.humaninteraction.UserTask;
import org.yaoqiang.bpmn.model.elements.process.BPMNProcess;
import org.yaoqiang.bpmn.model.elements.process.Lane;
import org.yaoqiang.bpmn.model.elements.process.LaneSet;
import org.yaoqiang.graph.util.Constants;

import com.mxgraph.io.mxCodec;
import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGeometry;
import com.mxgraph.model.mxGraphModel;
import com.mxgraph.util.mxConstants;
import com.mxgraph.util.mxDomUtils;
import com.mxgraph.util.mxPoint;
import com.mxgraph.util.mxRectangle;
import com.mxgraph.util.mxUtils;
import com.mxgraph.view.mxCellState;

/**
 * BPMNCodecUtils
 * 
 * @author Shi Yaoqiang(shi_yaoqiang@yahoo.com)
 */
public class BPMNCodecUtils {

	public static String generateNodeValue(Map<String, XMLElement> bpmnElementMap, XMLComplexElement bpmnElement) {
		if (bpmnElement instanceof TextAnnotation) {
			return bpmnElement.get("text").toValue();
		} else if (bpmnElement instanceof ItemAwareElement) {
			String name = bpmnElement.get("name").toValue();
			if (bpmnElement instanceof DataObjectReference) {
				DataObject dataObject = (DataObject) bpmnElementMap.get(((DataObjectReference) bpmnElement).getDataObjectRef());
				if (dataObject != null) {
					name = dataObject.getName();
				}
			} else if (bpmnElement instanceof DataStoreReference) {
				DataStore dataStore = (DataStore) bpmnElementMap.get(((DataStoreReference) bpmnElement).getDataStoreRef());
				if (dataStore != null) {
					name = dataStore.getName();
				}
			}
			String state = ((ItemAwareElement) bpmnElement).getDataState();
			if (state.length() != 0) {
				state = "\n[" + state + "]";
			}
			return name + state;
		} else if (bpmnElement instanceof Group) {
			CategoryValue categoryValue = (CategoryValue) bpmnElementMap.get(((Group) bpmnElement).getCategoryValueRef());
			if (categoryValue == null) {
				return "";
			}
			Category category = (Category) bpmnElementMap.get(((Category) categoryValue.getParent().getParent()).getId());
			if (category.getName().length() == 0) {
				return categoryValue.getValue();
			} else {
				return category.getName() + ":" + categoryValue.getValue();
			}
		} else {
			try {
				return bpmnElement.get("name").toValue();
			} catch (Exception e) {
				return "";
			}
		}
	}

	public static mxCell generateNodeParent(BPMNGraph graph, Map<String, XMLElement> bpmnElementMap, BPMNShape shape) {
		String parentId = null;
		XMLElement bpmnElement = bpmnElementMap.get(shape.getBpmnElement());
		if (bpmnElement instanceof BoundaryEvent) {
			parentId = ((BoundaryEvent) bpmnElement).getAttachedToRef();
		} else if (bpmnElement instanceof Participant) {
			String bandKind = shape.getParticipantBandKind();
			if (bandKind.length() != 0) {
				XMLElement pShape = bpmnElementMap.get(shape.getChoreographyActivityShape());
				parentId = ((BPMNShape) pShape).getBpmnElement() + "_act";
			}
		} else {
			XMLElement parent = bpmnElement.getParent().getParent();
			if (parent instanceof Collaboration) {

			} else if (parent instanceof LaneSet) {
				parent = parent.getParent();
				if (parent instanceof Lane) {
					parentId = ((Lane) parent).getId();
				} else {
					parent = parent.getParent();
					if (parent instanceof Lane) {
						parentId = ((Lane) parent).getId();
					} else if (parent instanceof BPMNProcess) {
						String processId = ((BPMNProcess) parent).getId();
						parent = parent.getParent().getParent();
						parentId = BPMNModelUtils.getParticipantByProcessId(processId, (Definitions) parent);
					}
				}
			} else if (parent instanceof BPMNProcess) {
				if (bpmnElement instanceof FlowElement) {
					String nodeId = ((FlowElement) bpmnElement).getId();
					if (((BPMNProcess) parent).getLanes().isEmpty()) {
						String processId = ((BPMNProcess) parent).getId();
						parent = parent.getParent().getParent();
						parentId = BPMNModelUtils.getParticipantByProcessId(processId, (Definitions) parent);
					} else {
						for (XMLElement lane : ((BPMNProcess) parent).getLanes()) {
							for (XMLElement flownode : ((Lane) lane).getFlowNodeList()) {
								if (nodeId.equals(flownode.toValue())) {
									parentId = ((Lane) lane).getId();
									break;
								}
							}
							if (parentId != null) {
								break;
							}
						}
					}
				}
			} else if (parent instanceof SubProcess) {
				parentId = ((SubProcess) parent).getId();
			} else if (parent instanceof SubChoreography) {
				parentId = ((SubChoreography) parent).getId();
			}
		}

		if (parentId == null || parentId.length() == 0 || graph.getModel().getCell(parentId) == null) {
			return (mxCell) graph.getModel().getChildAt(graph.getModel().getRoot(), 0);
		}
		return (mxCell) graph.getModel().getCell(parentId);
	}

	public static mxGeometry generateNodeGeometry(mxGraphModel model, mxCell parent, BPMNShape shape, XMLComplexElement bpmnElement) {
		Bounds bounds = shape.getBounds();
		if (bounds.getX() == -1 && bounds.getY() == -1 && !(bpmnElement instanceof BoundaryEvent)) {
			bounds.setX(20);
			bounds.setY(200);
		}
		if (bpmnElement instanceof Event) {
			String width = "";
			String height = "";
			if (bpmnElement instanceof StartEvent) {
				width = Constants.SETTINGS.getProperty("style_startEvent_size", "32");
				height = Constants.SETTINGS.getProperty("style_startEvent_size", "32");
			} else if (bpmnElement instanceof EndEvent) {
				width = Constants.SETTINGS.getProperty("style_endEvent_size", "32");
				height = Constants.SETTINGS.getProperty("style_endEvent_size", "32");
			} else {
				width = Constants.SETTINGS.getProperty("style_intermediateEvent_size", "32");
				height = Constants.SETTINGS.getProperty("style_intermediateEvent_size", "32");
			}
			bounds.setWidth(Integer.parseInt(width));
			bounds.setHeight(Integer.parseInt(height));
		} else if (bpmnElement instanceof Gateway) {
			String width = Constants.SETTINGS.getProperty("style_gateway_size", "42");
			String height = Constants.SETTINGS.getProperty("style_gateway_size", "42");
			bounds.setWidth(Integer.parseInt(width));
			bounds.setHeight(Integer.parseInt(height));
		} else if (bpmnElement instanceof DataObjectReference || bpmnElement instanceof DataInput || bpmnElement instanceof DataOutput) {
			String width = Constants.SETTINGS.getProperty("style_dataobject_width", "29");
			String height = Constants.SETTINGS.getProperty("style_dataobject_height", "38");
			bounds.setWidth(Integer.parseInt(width));
			bounds.setHeight(Integer.parseInt(height));
		} else if (bpmnElement instanceof DataStoreReference) {
			String width = Constants.SETTINGS.getProperty("style_datastore_width", "35");
			String height = Constants.SETTINGS.getProperty("style_datastore_height", "30");
			bounds.setWidth(Integer.parseInt(width));
			bounds.setHeight(Integer.parseInt(height));
		} else if (bpmnElement instanceof ConversationNode) {
			String width = Constants.SETTINGS.getProperty("style_conversation_width", "40");
			String height = Constants.SETTINGS.getProperty("style_conversation_height", "35");
			bounds.setWidth(Integer.parseInt(width));
			bounds.setHeight(Integer.parseInt(height));
		} else if (bpmnElement instanceof Activity) {
			String width = Constants.SETTINGS.getProperty("style_task_width", "85");
			String height = Constants.SETTINGS.getProperty("style_task_height", "55");
			if (bounds.getWidth() < Integer.parseInt(width)) {
				bounds.setWidth(Integer.parseInt(width));
			}
			if (bounds.getHeight() < Integer.parseInt(height)) {
				bounds.setHeight(Integer.parseInt(height));
			}
		} else if (bpmnElement instanceof Participant) {
			String bandKind = shape.getParticipantBandKind();
			if (bandKind.length() != 0) {
				if (((Participant) bpmnElement).getMultiplicity() > 1 || ((Participant) bpmnElement).getMultiplicity() == -1) {
					bounds.setHeight(Constants.PARTICIPANT_HEIGHT * 1.75);
				} else {
					bounds.setHeight(Constants.PARTICIPANT_HEIGHT);
				}
			}
		}

		mxGeometry pgeo = new mxGeometry();
		mxCell tmp = parent;
		while (tmp != null && tmp.getGeometry() != null) {
			pgeo.setX(pgeo.getX() + tmp.getGeometry().getX());
			pgeo.setY(pgeo.getY() + tmp.getGeometry().getY());
			tmp = (mxCell) model.getParent(tmp);
		}

		mxGeometry geometry = new mxGeometry(bounds.getX() - pgeo.getX() + 50, bounds.getY() - pgeo.getY() + 50, bounds.getWidth(), bounds.getHeight());
		if (bpmnElement instanceof SubProcess) {
			mxRectangle rect = new mxRectangle(geometry.getX(), geometry.getY(), 85, 55);
			if (!shape.isExpanded()) {
				rect.setWidth(400);
				rect.setHeight(250);
			}
			geometry.setAlternateBounds(rect);
		} else if (bpmnElement instanceof BoundaryEvent) {
			geometry.setOffset(new mxPoint(geometry.getX(), geometry.getY()));
			geometry.setX(0);
			geometry.setY(0);
			geometry.setRelative(true);
			if (bounds.getX() == -1 && bounds.getY() == -1) {
				geometry.setX(1);
				geometry.setY(1);
				geometry.setOffset(new mxPoint(-parent.getGeometry().getWidth() * 0.2 - 32, -32 / 2));
			}
		}

		return geometry;
	}

	public static String generateNodeStyle(BPMNGraph graph, mxCell parent, Map<String, XMLElement> bpmnElementMap, BPMNShape shape) {
		String style = "";
		XMLElement bpmnElement = bpmnElementMap.get(shape.getBpmnElement());
		if (graph.isChoreography(parent)) {
			parent = (mxCell) parent.getParent();
		}
		if (shape.getBPMNLabel().getBounds().getWidth() == 0 && !graph.isInCollapsedSwimlane(parent) && !graph.isInCollapsedSubProcess(parent)
				&& !graph.isCallActivityProcess(parent)) {
			style += "noLabel=1;";
		}
		Element node = mxDomUtils.createDocument().createElement("node");
		if (bpmnElement instanceof Lane) {
			String collapsed = ";horizontal=0";
			if (shape.isHorizontal()) {
				if (!shape.isExpanded()) {
					collapsed = ";horizontal=1";
				}
				style += "lane;whiteSpace=wrap" + collapsed;
			} else {
				if (shape.isExpanded()) {
					collapsed = ";horizontal=1";
				}
				style += "lane;type=vertical;whiteSpace=wrap" + collapsed;
			}
		} else if (bpmnElement instanceof Activity) {
			if (Constants.SETTINGS.getProperty("labelWrap", "1").equals("1")) {
				style += "whiteSpace=wrap;";
			}

			if (((Activity) bpmnElement).isForCompensation()) {
				style += "compensation=1;compensationImage=/org/yaoqiang/graph/images/marker_compensation.png;";
			}

			if (bpmnElement instanceof Task) {
				if (bpmnElement instanceof BusinessRuleTask) {
					style += "taskBusinessRule;";
				} else if (bpmnElement instanceof ReceiveTask) {
					style += "taskReceive;";
				} else if (bpmnElement instanceof UserTask) {
					style += "taskUser;";
				} else if (bpmnElement instanceof ManualTask) {
					style += "taskManual;";
				} else if (bpmnElement instanceof SendTask) {
					style += "taskSend;";
				} else if (bpmnElement instanceof ScriptTask) {
					style += "taskScript;";
				} else if (bpmnElement instanceof ServiceTask) {
					style += "taskService;";
				} else {
					style += "task;";
				}
			} else if (bpmnElement instanceof SubProcess) {
				if (bpmnElement instanceof Transaction) {
					style += "tranSubprocess;";
				} else if (bpmnElement instanceof AdHocSubProcess) {
					style += "adHocSubprocess;";
				} else {
					if (((SubProcess) bpmnElement).isTriggeredByEvent()) {
						style += "eventSubprocess;";
					} else {
						style += "subprocess;";
					}
				}
			} else if (bpmnElement instanceof CallActivity) {
				String calledElementId = ((CallActivity) bpmnElement).getCalledElement();
				CallableElement calledElement = (CallableElement) bpmnElementMap.get(calledElementId);
				if (calledElement != null) {
					node.setAttribute("call", calledElement.getName().length() == 0 ? calledElementId : calledElement.getName());
				}
				if (calledElement == null || calledElement instanceof BPMNProcess) {
					style += "callSubprocess;";
				} else if (calledElement instanceof GlobalBusinessRuleTask) {
					style += "callBusinessRule;";
				} else if (calledElement instanceof GlobalScriptTask) {
					style += "callScript;";
				} else if (calledElement instanceof GlobalManualTask) {
					style += "callManual;";
				} else if (calledElement instanceof GlobalUserTask) {
					style += "callUser;";
				} else {
					style += "callActivity;";
				}
			}

			LoopCharacteristics loopType = ((Activity) bpmnElement).getLoopCharacteristics();
			if (loopType == null) {
			} else if (loopType instanceof MultiInstanceLoopCharacteristics) {
				if (((MultiInstanceLoopCharacteristics) loopType).isSequential()) {
					style += "loop=multi_instance_sequential;loopImage=/org/yaoqiang/graph/images/marker_multiple_sequential.png;";
				} else {
					style += "loop=multi_instance;loopImage=/org/yaoqiang/graph/images/marker_multiple.png;";
				}
			} else {
				style += "loop=standard;loopImage=/org/yaoqiang/graph/images/marker_loop.png;";
			}
		} else if (bpmnElement instanceof Gateway) {
			style += "";
			if (bpmnElement instanceof ComplexGateway) {
				style += "complexGateway;";
			} else if (bpmnElement instanceof EventBasedGateway) {
				if (((EventBasedGateway) bpmnElement).isInstantiate()) {
					String type = ((EventBasedGateway) bpmnElement).getEventGatewayType();
					if (type.equals("Exclusive")) {
						style += "eventGatewayInstantiate;";
					} else if (type.equals("Parallel")) {
						style += "parallelEventGateway;";
					}
				} else {
					style += "eventGateway;";
				}
			} else if (bpmnElement instanceof InclusiveGateway) {
				style += "inclusiveGateway;";
			} else if (bpmnElement instanceof ParallelGateway) {
				style += "parallelGateway;";
			} else if (bpmnElement instanceof ExclusiveGateway) {
				if (shape.isMarkerVisible()) {
					style += "exclusiveGatewayWithIndicator;";
				} else {
					style += "exclusiveGateway;";
				}
			} else {
				style += "gateway;";
			}
		} else if (bpmnElement instanceof Event) {
			if (bpmnElement instanceof CatchEvent) {
				if (bpmnElement instanceof StartEvent) {
					StartEvent event = (StartEvent) bpmnElement;
					List<XMLElement> eventDefs = event.getEventDefinitionList();
					eventDefs.addAll(event.getRefEventDefinitionList());
					if (eventDefs.isEmpty()) {
						style += "startEvent;";
					} else if (eventDefs.size() > 1) {
						if (event.isParallelMultiple()) {
							if (event.isInterrupting()) {
								style += "startParallelMultipleEvent;";
							} else {
								style += "startParallelMultipleNonInterruptingEvent;";
							}
						} else {
							if (event.isInterrupting()) {
								style += "startMultipleEvent;";
							} else {
								style += "startMultipleNonInterruptingEvent;";
							}
						}
					} else {
						XMLElement eventDef = eventDefs.get(0);
						if (event.isInterrupting()) {
							if (eventDef instanceof MessageEventDefinition) {
								style += "startMessageEvent;";
							} else if (eventDef instanceof TimerEventDefinition) {
								style += "startTimerEvent;";
							} else if (eventDef instanceof EscalationEventDefinition) {
								style += "startEscalationEvent;";
							} else if (eventDef instanceof ConditionalEventDefinition) {
								style += "startConditionalEvent;";
							} else if (eventDef instanceof ErrorEventDefinition) {
								style += "startErrorEvent;";
							} else if (eventDef instanceof CompensateEventDefinition) {
								style += "startCompensationEvent;";
							} else if (eventDef instanceof SignalEventDefinition) {
								style += "startSignalEvent;";
							}
						} else {
							if (eventDef instanceof MessageEventDefinition) {
								style += "startMessageNonInterruptingEvent;";
							} else if (eventDef instanceof TimerEventDefinition) {
								style += "startTimerNonInterruptingEvent;";
							} else if (eventDef instanceof EscalationEventDefinition) {
								style += "startEscalationNonInterruptingEvent;";
							} else if (eventDef instanceof ConditionalEventDefinition) {
								style += "startConditionalNonInterruptingEvent;";
							} else if (eventDef instanceof SignalEventDefinition) {
								style += "startSignalNonInterruptingEvent;";
							}
						}
					}
				} else if (bpmnElement instanceof IntermediateCatchEvent) {
					IntermediateCatchEvent event = (IntermediateCatchEvent) bpmnElement;
					List<XMLElement> eventDefs = event.getEventDefinitionList();
					eventDefs.addAll(event.getRefEventDefinitionList());
					if (eventDefs.isEmpty()) {
						style += "intermediateEvent;";
					} else if (eventDefs.size() > 1) {
						if (event.isParallelMultiple()) {
							style += "intermediateParallelMultipleEvent;";
						} else {
							style += "intermediateMultipleEvent;";
						}
					} else {
						XMLElement eventDef = eventDefs.get(0);
						if (eventDef instanceof MessageEventDefinition) {
							style += "intermediateMessageEvent;";
						} else if (eventDef instanceof TimerEventDefinition) {
							style += "intermediateTimerEvent;";
						} else if (eventDef instanceof ConditionalEventDefinition) {
							style += "intermediateConditionalEvent;";
						} else if (eventDef instanceof LinkEventDefinition) {
							style += "intermediateLinkEvent;";
						} else if (eventDef instanceof SignalEventDefinition) {
							style += "intermediateSignalEvent;";
						}
					}
				} else if (bpmnElement instanceof BoundaryEvent) {
					BoundaryEvent event = (BoundaryEvent) bpmnElement;
					node.setAttribute("attached", event.getAttachedToRef());
					List<XMLElement> eventDefs = event.getEventDefinitionList();
					eventDefs.addAll(event.getRefEventDefinitionList());
					if (eventDefs.isEmpty()) {
						style += "intermediateEvent;";
					} else if (eventDefs.size() > 1) {
						if (event.isParallelMultiple()) {
							if (event.isCancelActivity()) {
								style += "intermediateParallelMultipleEvent;";
							} else {
								style += "intermediateParallelMultipleNonInterruptingEvent;";
							}
						} else {
							if (event.isCancelActivity()) {
								style += "intermediateMultipleEvent;";
							} else {
								style += "intermediateMultipleNonInterruptingEvent;";
							}
						}
					} else {
						XMLElement eventDef = eventDefs.get(0);
						if (event.isCancelActivity()) {
							if (eventDef instanceof MessageEventDefinition) {
								style += "intermediateMessageEvent;";
							} else if (eventDef instanceof TimerEventDefinition) {
								style += "intermediateTimerEvent;";
							} else if (eventDef instanceof EscalationEventDefinition) {
								style += "intermediateEscalationEvent;";
							} else if (eventDef instanceof ConditionalEventDefinition) {
								style += "intermediateConditionalEvent;";
							} else if (eventDef instanceof ErrorEventDefinition) {
								style += "intermediateErrorEvent;";
							} else if (eventDef instanceof CancelEventDefinition) {
								style += "intermediateCancelEvent;";
							} else if (eventDef instanceof CompensateEventDefinition) {
								style += "intermediateCompensationEvent;";
							} else if (eventDef instanceof SignalEventDefinition) {
								style += "intermediateSignalEvent;";
							}
						} else {
							if (eventDef instanceof MessageEventDefinition) {
								style += "intermediateMessageNonInterruptingEvent;";
							} else if (eventDef instanceof TimerEventDefinition) {
								style += "intermediateTimerNonInterruptingEvent;";
							} else if (eventDef instanceof EscalationEventDefinition) {
								style += "intermediateEscalationNonInterruptingEvent;";
							} else if (eventDef instanceof ConditionalEventDefinition) {
								style += "intermediateConditionalNonInterruptingEvent;";
							} else if (eventDef instanceof SignalEventDefinition) {
								style += "intermediateSignalNonInterruptingEvent;";
							}
						}
					}
				}
			} else if (bpmnElement instanceof ThrowEvent) {
				if (bpmnElement instanceof EndEvent) {
					EndEvent event = (EndEvent) bpmnElement;
					List<XMLElement> eventDefs = event.getEventDefinitionList();
					eventDefs.addAll(event.getRefEventDefinitionList());
					if (eventDefs.isEmpty()) {
						style += "endEvent;";
					} else if (eventDefs.size() > 1) {
						style += "endMultipleEvent;";
					} else {
						XMLElement eventDef = eventDefs.get(0);
						if (eventDef instanceof MessageEventDefinition) {
							style += "endMessageEvent;";
						} else if (eventDef instanceof EscalationEventDefinition) {
							style += "endEscalationEvent;";
						} else if (eventDef instanceof ErrorEventDefinition) {
							style += "endErrorEvent;";
						} else if (eventDef instanceof CancelEventDefinition) {
							style += "endCancelEvent;";
						} else if (eventDef instanceof CompensateEventDefinition) {
							style += "endCompensationEvent;";
						} else if (eventDef instanceof SignalEventDefinition) {
							style += "endSignalEvent;";
						} else if (eventDef instanceof TerminateEventDefinition) {
							style += "endTerminateEvent;";
						}
					}
				} else if (bpmnElement instanceof IntermediateThrowEvent) {
					IntermediateThrowEvent event = (IntermediateThrowEvent) bpmnElement;
					List<XMLElement> eventDefs = event.getEventDefinitionList();
					if (eventDefs.isEmpty()) {
						style += "intermediateEvent;";
					} else if (eventDefs.size() > 1) {
						style += "intermediateMultipleThrowEvent;";
					} else {
						XMLElement eventDef = eventDefs.get(0);
						if (eventDef instanceof MessageEventDefinition) {
							style += "intermediateMessageThrowEvent;";
						} else if (eventDef instanceof EscalationEventDefinition) {
							style += "intermediateEscalationThrowEvent;";
						} else if (eventDef instanceof LinkEventDefinition) {
							style += "intermediateLinkThrowEvent;";
						} else if (eventDef instanceof CompensateEventDefinition) {
							style += "intermediateCompensationThrowEvent;";
						} else if (eventDef instanceof SignalEventDefinition) {
							style += "intermediateSignalThrowEvent;";
						}
					}
				}
			}
		} else if (bpmnElement instanceof ConversationNode) {
			if (bpmnElement instanceof Conversation) {
				style += "conversation;";
			} else if (bpmnElement instanceof SubConversation) {
				style += "subConversation;";
			} else if (bpmnElement instanceof CallConversation) {
				XMLElement collaboration = bpmnElementMap.get(((CallConversation) bpmnElement).getCalledCollaboratioinRef());
				if (collaboration instanceof GlobalConversation) {
					style += "callConversation;";
				} else if (collaboration == null || collaboration instanceof Collaboration) {
					style += "callCollaboration;";
				}
			}
		} else if (bpmnElement instanceof ChoreographyActivity) {
			String loopType = ((ChoreographyActivity) bpmnElement).getLoopType();
			if (bpmnElement instanceof ChoreographyTask) {
				style += "choreography;";
			} else if (bpmnElement instanceof SubChoreography) {
				style += "subChoreography;";
			} else if (bpmnElement instanceof CallChoreography) {
				CallChoreography callC = (CallChoreography) bpmnElement;
				XMLElement choreography = bpmnElementMap.get(callC.getCalledChoreographyRef());
				if (choreography == null && callC.getId().endsWith("_CT") || choreography instanceof GlobalChoreographyTask) {
					style += "callChoreographyTask;";
				} else if (choreography == null || choreography instanceof Choreography) {
					style += "callChoreography;";
				}
			}
			if (loopType.equals("Standard")) {
				style += "loop=standard;loopImage=/org/yaoqiang/graph/images/marker_loop.png;";
			} else if (loopType.equals("MultiInstanceParallel")) {
				style += "loop=multi_instance;loopImage=/org/yaoqiang/graph/images/marker_multiple.png;";
			} else if (loopType.equals("MultiInstanceSequential")) {
				style += "loop=multi_instance_sequential;loopImage=/org/yaoqiang/graph/images/marker_multiple_sequential.png;";
			}
		} else if (bpmnElement instanceof Participant) {
			BPMNProcess processElement = (BPMNProcess) bpmnElementMap.get(((Participant) bpmnElement).getProcessRef());
			if (processElement != null) {
				node.setAttribute("process", processElement.getName().length() == 0 ? processElement.getId() : processElement.getName());
			}

			String bandKind = shape.getParticipantBandKind();
			Bounds bounds = shape.getBounds();
			XMLElement pShape = bpmnElementMap.get(shape.getChoreographyActivityShape());
			if (((Participant) bpmnElement).getMultiplicity() > 1 || ((Participant) bpmnElement).getMultiplicity() == -1) {
				style += "multi_instance=1;";
			}
			if (bandKind.length() == 0) {
				String collapsed = ";horizontal=0";
				if (shape.isHorizontal()) {
					if (!shape.isExpanded()) {
						collapsed = ";horizontal=1";
					}
					style += "swimlane;auto=0;whiteSpace=wrap" + collapsed;
				} else {
					if (shape.isExpanded()) {
						collapsed = ";horizontal=1";
					}
					style += "swimlane;auto=0;type=vertical;whiteSpace=wrap" + collapsed;
				}
			} else if (pShape != null && bandKind.length() != 0) {
				Object act = graph.getModel().getCell(((BPMNShape) pShape).getBpmnElement());
				if (graph.isCallChoreographyActivity(act)) {
					style += "call=1;strokeWidth=3;";
				}
				mxGeometry ageo = graph.getModel().getGeometry(act);
				ageo.setHeight(ageo.getHeight() - bounds.getHeight());
				if (bandKind.equals("middle_initiating")) {
					if (bounds.getY() > ageo.getY()) {
						style += "participantAdditionalBottom;fillColor=#E8EEF7";
					} else {
						ageo.setY(ageo.getY() + bounds.getHeight());
						style += "participantAdditionalTop;fillColor=#E8EEF7";
					}
				} else if (bandKind.equals("middle_non_initiating")) {
					if (bounds.getY() > ageo.getY()) {
						style += "participantAdditionalBottom";
					} else {
						ageo.setY(ageo.getY() + bounds.getHeight());
						style += "participantAdditionalTop";
					}
				} else if (bandKind.equals("top_initiating")) {
					ageo.setY(ageo.getY() + bounds.getHeight());
					style += "participantTop;fillColor=#E8EEF7";
				} else if (bandKind.equals("top_non_initiating")) {
					ageo.setY(ageo.getY() + bounds.getHeight());
					style += "participantTop";
				} else if (bandKind.equals("bottom_initiating")) {
					style += "participantBottom;fillColor=#E8EEF7";
				} else if (bandKind.equals("bottom_non_initiating")) {
					style += "participantBottom";
				}
				graph.getModel().setGeometry(act, ageo);
			}
		} else if (bpmnElement instanceof TextAnnotation) {
			style += "annotation;whiteSpace=wrap;";
		} else if (bpmnElement instanceof Message) {
			style += "initiatingMessage;";
		} else if (bpmnElement instanceof Group) {
			style += "group;";
		} else if (bpmnElement instanceof DataStoreReference) {
			style += "datastore;";
		} else if (bpmnElement instanceof DataObjectReference) {
			DataObjectReference dataObjectReference = (DataObjectReference) bpmnElement;
			DataObject dataObject = (DataObject) bpmnElementMap.get(dataObjectReference.getDataObjectRef());
			if (dataObject.isCollection()) {
				style += "dataobjects;";
			} else {
				style += "dataobject;";
			}
		} else if (bpmnElement instanceof DataInput) {
			if (((DataInput) bpmnElement).isCollection()) {
				style += "datainputs;";
			} else {
				style += "datainput;";
			}
		} else if (bpmnElement instanceof DataOutput) {
			if (((DataOutput) bpmnElement).isCollection()) {
				style += "dataoutputs;";
			} else {
				style += "dataoutput;";
			}
		}
		return style;
	}

	public static String generateEdgeStyle(Map<String, XMLElement> bpmnElementMap, XMLElement el) {
		String style = "";
		XMLComplexElement bpmnElement = (XMLComplexElement) bpmnElementMap.get(((BPMNEdge) el).getBpmnElement());

		if (bpmnElement instanceof SequenceFlow) {
			SequenceFlow sequenceFlow = (SequenceFlow) bpmnElement;
			String expression = sequenceFlow.getConditionExpression().toValue();
			if (expression.length() == 0) {
				XMLComplexElement srcNode = (XMLComplexElement) bpmnElementMap.get(sequenceFlow.getSourceRef());
				if (srcNode == null) {
					return style;
				}
				XMLElement defaultFlow = srcNode.get("default");
				if (defaultFlow != null && defaultFlow.toValue().equals(sequenceFlow.getId())) {
					style = Constants.EDGE_TYPE_DEFAULT_SEQUENCE_FLOW;
				} else {
					style = Constants.EDGE_TYPE_SEQUENCE_FLOW;
				}
			} else {
				if (bpmnElementMap.get(sequenceFlow.getSourceRef()) instanceof Gateway) {
					style = Constants.EDGE_TYPE_SEQUENCE_FLOW + ";expression=" + expression;
				} else {
					style = Constants.EDGE_TYPE_CONDITION_SEQUENCE_FLOW + ";expression=" + expression;
				}
			}
		} else if (bpmnElement instanceof MessageFlow) {
			style = Constants.EDGE_TYPE_MESSAGE_FLOW;
		} else if (bpmnElement instanceof DataAssociation) {
			style = Constants.EDGE_TYPE_DATA_ASSOCIATION;
		} else if (bpmnElement instanceof Association) {
			if (bpmnElementMap.get(bpmnElement.get("sourceRef").toValue()) instanceof BoundaryEvent) {
				style = Constants.EDGE_TYPE_COMPENSATION_ASSOCIATION;
			} else {
				style = Constants.EDGE_TYPE_ASSOCIATION;
			}
		} else if (bpmnElement instanceof ConversationLink) {
			style = Constants.EDGE_TYPE_CONVERSATION_LINK;
		}

		return style;
	}

	public static String generateEdgeValue(Map<String, XMLElement> bpmnElementMap, XMLElement el) {
		XMLComplexElement bpmnElement = (XMLComplexElement) bpmnElementMap.get(((BPMNEdge) el).getBpmnElement());
		return bpmnElement instanceof Association || bpmnElement instanceof DataAssociation ? "" : bpmnElement.get("name").toValue();
	}

	public static mxCell generateEdgeTerminal(mxGraphModel model, XMLComplexElement bpmnElement, BPMNEdge edge, boolean isSource) {
		String sourceId = bpmnElement.get("sourceRef").toValue();
		String targetId = bpmnElement.get("targetRef").toValue();
		if (bpmnElement instanceof DataInputAssociation) {
			sourceId = ((DataInputAssociation) bpmnElement).getSourceRef();
			FlowElement target = (FlowElement) bpmnElement.getParent().getParent();
			targetId = target.getId();
		} else if (bpmnElement instanceof DataOutputAssociation) {
			FlowElement source = (FlowElement) bpmnElement.getParent().getParent();
			sourceId = source.getId();
			targetId = ((DataOutputAssociation) bpmnElement).getTargetRef();
		}
		if (sourceId.equals("_1") || sourceId.equals("_0")) {
			sourceId = "_" + sourceId;
			bpmnElement.set("sourceRef", sourceId);
		}
		if (targetId.equals("_1") || targetId.equals("_0")) {
			targetId = "_" + targetId;
			bpmnElement.set("targetRef", targetId);
		}
		if (isSource) {
			return (mxCell) model.getCell(sourceId);
		} else {
			return (mxCell) model.getCell(targetId);
		}
	}

	public static Object insertChoreography(mxCodec codec, mxGraphModel model, String loopType, String id, mxCell parent, Object value, mxGeometry geometry,
			String style) {
		if (loopType.equals("Standard")) {
			style += ";loop=standard;loopImage=/org/yaoqiang/graph/images/marker_loop.png";
		} else if (loopType.equals("MultiInstanceSequential")) {
			style += ";loop=multi_instance;loopImage=/org/yaoqiang/graph/images/marker_multiple.png";
		} else if (loopType.equals("MultiInstanceParallel")) {
			style += ";loop=multi_instance_sequential;loopImage=/org/yaoqiang/graph/images/marker_multiple_sequential.png";
		}
		parent.setValue(null);

		mxCell cell = new mxCell(value, geometry, style);
		cell.setId(id);
		cell.setVertex(true);
		cell.setParent(parent);
		codec.insertIntoGraph(cell);
		model.getCells().put(cell.getId(), cell);
		return cell;
	}

	public static List<mxPoint> convertTomxPointList(List<Point> points, mxGeometry geo) {
		List<mxPoint> mxPointList = new ArrayList<mxPoint>();
		if (points == null) {
			return mxPointList;
		} else {
			if (!points.isEmpty()) {
				for (int i = 0; i < points.size(); i++) {
					Point p = points.get(i);
					mxPointList.add(new mxPoint(p.getX() - geo.getX() + 50, p.getY() - geo.getY() + 50));
				}
			}
		}
		return mxPointList;
	}

	public static void setBounds(BPMNGraph graph, mxCell cell, Element bounds) {
		mxGraphModel model = graph.getModel();
		mxCell pCell = (mxCell) model.getParent(cell);
		mxGeometry pgeo = new mxGeometry();
		while (pCell != null && pCell.getGeometry() != null) {
			pgeo.setX(pgeo.getX() + pCell.getGeometry().getX());
			pgeo.setY(pgeo.getY() + pCell.getGeometry().getY());
			pCell = (mxCell) model.getParent(pCell);
		}

		mxGeometry geo = (mxGeometry) model.getGeometry(cell).clone();
		if (graph.isAttachedEvent(cell)) {
			geo.setX(pgeo.getX() - 50);
			geo.setY(pgeo.getY() - 50);
		} else if (graph.isAttachedMessage(cell)) {
			mxPoint offset = geo.getOffset();
			geo.setX(offset.getX());
			geo.setY(offset.getY());
		} else {
			geo.setX(geo.getX() + pgeo.getX() - 50);
			geo.setY(geo.getY() + pgeo.getY() - 50);
		}

		bounds.setAttribute("height", Double.toString(geo.getHeight()));
		bounds.setAttribute("width", Double.toString(geo.getWidth()));
		if (graph.isAttachedEvent(cell)) {
			mxGeometry cellgeo = model.getGeometry(cell);
			mxGeometry parentgeo = model.getGeometry(model.getParent(cell));
			mxPoint offset = cellgeo.getOffset();
			if (cellgeo.getX() == 0) {
				geo.setX(geo.getX() + offset.getX());
			} else if (cellgeo.getX() == 1) {
				geo.setX(geo.getX() + parentgeo.getWidth() + offset.getX());
			}
			if (cellgeo.getY() == 0) {
				geo.setY(geo.getY() + offset.getY());
			} else if (cellgeo.getY() == 1) {
				geo.setY(geo.getY() + parentgeo.getHeight() + offset.getY());
			}
			bounds.setAttribute("x", Double.toString(geo.getX()));
			bounds.setAttribute("y", Double.toString(geo.getY()));
		} else {
			bounds.setAttribute("x", Double.toString(geo.getX()));
			bounds.setAttribute("y", Double.toString(geo.getY()));
		}
	}

	public static void setLabelBounds(BPMNGraph graph, mxCell cell, Element bounds) {
		mxRectangle labelBounds = new mxRectangle();
		double scale = graph.getView().getScale();
		mxPoint translate = graph.getView().getTranslate();
		mxCellState state = graph.getView().getState(cell);
		if (state != null && state.getLabelBounds() != null && !mxUtils.isTrue(state.getStyle(), mxConstants.STYLE_NOLABEL, false)) {
			labelBounds = state.getLabelBounds();
		}
		bounds.setAttribute("x", Double.toString(labelBounds.getX() == 0 ? 0 : labelBounds.getX() / scale - translate.getX()));
		bounds.setAttribute("y", Double.toString(labelBounds.getY() == 0 ? 0 : labelBounds.getY() / scale - translate.getY()));
		bounds.setAttribute("height", Double.toString(labelBounds.getHeight() / scale));
		bounds.setAttribute("width", Double.toString(labelBounds.getWidth() / scale));
	}

	public static Element generateEdgeElement(BPMNGraph graph, mxCell cell, Element plane, String bpmndiPrefix, String dcPrefix, String diPrefix) {
		Element el = plane.getOwnerDocument().createElement(bpmndiPrefix + "BPMNEdge");
		Element label = plane.getOwnerDocument().createElement(bpmndiPrefix + "BPMNLabel");
		Element bounds = plane.getOwnerDocument().createElement(dcPrefix + "Bounds");

		mxGeometry geo = graph.getModel().getGeometry(cell);
		mxPoint offset = geo.getOffset();
		if (offset == null) {
			offset = new mxPoint();
		}

		bounds.setAttribute("x", Double.toString(geo.getX()));
		bounds.setAttribute("y", Double.toString(geo.getY()));
		bounds.setAttribute("width", Double.toString(offset.getX()));
		bounds.setAttribute("height", Double.toString(offset.getY()));
		label.appendChild(bounds);

		BPMNCodecUtils.generateWaypoints(diPrefix, el, graph, cell);
		el.appendChild(label);
		String id = cell.getId();
		// if (id.startsWith("__1")) {
		// id = id.substring(1);
		// }
		String sid = cell.getSource().getId();
		// if (sid.startsWith("__1")) {
		// sid = sid.substring(1);
		// }
		String tid = cell.getTarget().getId();
		// if (tid.startsWith("__1")) {
		// tid = tid.substring(1);
		// }
		el.setAttribute("id", "Yaoqiang-" + id);
		el.setAttribute("bpmnElement", id);
		el.setAttribute("sourceElement", sid);
		el.setAttribute("targetElement", tid);
		return el;
	}

	public static void generateWaypoints(String prefix, Element edge, BPMNGraph graph, mxCell cell) {
		mxGeometry geo = graph.getModel().getGeometry(cell);
		mxCell pCell = (mxCell) graph.getModel().getParent(cell);
		mxGeometry pgeo = new mxGeometry();
		while (pCell != null && pCell.getGeometry() != null) {
			pgeo.setX(pgeo.getX() + pCell.getGeometry().getX());
			pgeo.setY(pgeo.getY() + pCell.getGeometry().getY());
			pCell = (mxCell) graph.getModel().getParent(pCell);
		}

		mxCellState state = graph.getView().getState(cell);
		List<mxPoint> abps = null;
		double scale = graph.getView().getScale();
		List<mxPoint> points = new ArrayList<mxPoint>();
		if (state == null) {
			Object srcCell = graph.getModel().getTerminal(cell, true);
			mxGeometry srcGeo = graph.getModel().getGeometry(srcCell);
			mxPoint src = new mxPoint(srcGeo.getCenterX() + pgeo.getX() - 50, srcGeo.getCenterY() + pgeo.getY() - 50);
			points.add(src);
		} else {
			abps = state.getAbsolutePoints();
			mxPoint src = (mxPoint) abps.get(0).clone();
			src.setX(src.getX() / scale - graph.getView().getTranslate().getX() - 50);
			src.setY(src.getY() / scale - graph.getView().getTranslate().getY() - 50);
			points.add(src);
		}

		if (geo.getPoints() != null) {
			mxPoint cp = null;
			for (mxPoint p : geo.getPoints()) {
				cp = new mxPoint(p.getX() + pgeo.getX() - 50, p.getY() + pgeo.getY() - 50);
				points.add(cp);
			}
		}

		if (state == null) {
			Object tgtCell = graph.getModel().getTerminal(cell, false);
			mxGeometry tgtGeo = graph.getModel().getGeometry(tgtCell);
			mxPoint tgt = new mxPoint(tgtGeo.getCenterX() + pgeo.getX() - 50, tgtGeo.getCenterY() + pgeo.getY() - 50);
			points.add(tgt);
		} else {
			mxPoint tgt = (mxPoint) abps.get(abps.size() - 1).clone();
			tgt.setX(tgt.getX() / scale - graph.getView().getTranslate().getX() - 50);
			tgt.setY(tgt.getY() / scale - graph.getView().getTranslate().getY() - 50);
			points.add(tgt);
		}
		Element waypoint = null;
		for (mxPoint p : points) {
			waypoint = edge.getOwnerDocument().createElement(prefix + "waypoint");
			generateWaypoint(edge, waypoint, p);
		}
	}

	public static void generateWaypoint(Element edge, Element waypoint, mxPoint point) {
		waypoint.setAttribute("x", Double.toString(point.getX()));
		waypoint.setAttribute("y", Double.toString(point.getY()));
		edge.appendChild(waypoint);
	}

	public static void moveDiagram(BPMNDiagram diagram, Point offset) {
		List<XMLElement> shapes = diagram.getBPMNPlane().getBPMNShapes();
		for (XMLElement s : shapes) {
			BPMNShape shape = (BPMNShape) s;
			moveShape(shape, offset);
		}

		List<XMLElement> edges = diagram.getBPMNPlane().getBPMNEdges();
		for (XMLElement e : edges) {
			BPMNEdge edge = (BPMNEdge) e;
			moveEdge(edge, offset);
		}
	}

	public static void moveShape(BPMNShape shape, Point offset) {
		Bounds bounds = shape.getBounds();
		bounds.setX(bounds.getX() + offset.getX());
		bounds.setY(bounds.getY() + offset.getY());
		bounds = shape.getBPMNLabel().getBounds();
		bounds.setX(bounds.getX() + offset.getX());
		bounds.setY(bounds.getY() + offset.getY());
	}

	public static void moveEdge(BPMNEdge edge, Point offset) {
		List<XMLElement> waypoints = ((XMLCollection) edge.get("Waypoints")).getXMLElements();
		for (XMLElement wp : waypoints) {
			Waypoint waypoint = (Waypoint) wp;
			waypoint.setX(waypoint.getX() + offset.getX());
			waypoint.setY(waypoint.getY() + offset.getY());
		}
	}

	public static void mergeModel(BPMNGraph graph, Definitions fragment, Map<String, XMLElement> fragElementMap) {
		Definitions bpmnModel = graph.getBpmnModel();
		Map<String, XMLElement> bpmnElementMap = graph.getBpmnElementMap();
		Map<String, XMLElement> tempElementMap = new HashMap<String, XMLElement>();
		BPMNDiagram diagram = fragment.getFirstBPMNDiagram();
		if (!fragElementMap.isEmpty()) {
			int num = 1;
			boolean ok = false;
			while (!ok) {
				boolean has = false;
				for (String id : fragElementMap.keySet()) {
					String newId = "F" + num + id;
					if (bpmnElementMap.containsKey(newId)) {
						has = true;
						break;
					}
				}
				if (has) {
					num++;
				} else {
					ok = true;
				}
			}
			String prefix = "F" + num;

			for (XMLElement el : fragElementMap.values()) {
				XMLComplexElement bpmnEl = (XMLComplexElement) el;
				String id = bpmnEl.get("id").toValue();
				XMLComplexElement shapeEl = diagram.getShapeByBpmnElement(id);
				if (shapeEl != null) {
					shapeEl.set("bpmnElement", prefix + id);
				}
				if (shapeEl instanceof BPMNEdge) {
					shapeEl.set("sourceElement", prefix + shapeEl.get("sourceElement").toValue());
					shapeEl.set("targetElement", prefix + shapeEl.get("targetElement").toValue());
					if (bpmnEl instanceof SequenceFlow) {
						SequenceFlow sequenceFlow = (SequenceFlow) bpmnEl;
						sequenceFlow.setSourceRef(prefix + sequenceFlow.getSourceRef());
						sequenceFlow.setTargetRef(prefix + sequenceFlow.getTargetRef());
					} else if (bpmnEl instanceof Association) {
						Association association = (Association) bpmnEl;
						association.setSourceRef(prefix + association.getSourceRef());
						association.setTargetRef(prefix + association.getTargetRef());
					} else {
						// TODO:
					}
				} else {
					if (bpmnEl instanceof FlowNode) {
						FlowNode flowNode = (FlowNode) bpmnEl;
						for (XMLElement incoming : flowNode.getIncomings().getXMLElements()) {
							incoming.setValue(prefix + incoming.toValue());
						}
						for (XMLElement outgoing : flowNode.getOutgoings().getXMLElements()) {
							outgoing.setValue(prefix + outgoing.toValue());
						}
						if (flowNode instanceof Activity || flowNode instanceof ExclusiveGateway || flowNode instanceof InclusiveGateway
								|| flowNode instanceof ComplexGateway) {
							if (flowNode.get("default").toValue().length() != 0) {
								flowNode.set("default", prefix + flowNode.get("default").toValue());
							}
						} else if (flowNode instanceof ThrowEvent) {
							for (XMLElement ed : ((ThrowEvent) flowNode).getEventDefinitionList()) {
								if (ed instanceof CompensateEventDefinition) {
									CompensateEventDefinition ced = (CompensateEventDefinition) ed;
									if (ced.getActivityRef().length() != 0) {
										ced.setActivityRef(prefix + ced.getActivityRef());
									}
								}
							}
						} else if (flowNode instanceof BoundaryEvent) {
							BoundaryEvent event = (BoundaryEvent) flowNode;
							event.setAttachedToRef(prefix + event.getAttachedToRef());
						}
					}
				}
				bpmnEl.set("id", prefix + id);
				if (bpmnEl.getParent() instanceof XMLCollection) {
					((XMLCollection) bpmnEl.getParent()).remove(id);
					((XMLCollection) bpmnEl.getParent()).add(bpmnEl);
				}
				tempElementMap.put(prefix + id, bpmnEl);
				bpmnElementMap.put(prefix + id, bpmnEl);
			}

			// TODO:
			BPMNProcess fragProcess = BPMNModelUtils.getDefaultProcess(fragment);
			BPMNProcess process = BPMNModelUtils.getDefaultProcess(bpmnModel);
			if (fragProcess != null) {
				if (process == null) {
					bpmnModel.addRootElement(fragProcess);
				} else {
					process.addFlowElements(fragProcess.getFlowElements());
					process.addArtifacts(fragProcess.getArtifacts());
				}
			}

		}
	}

	public static void generateInterfaceForWSDL(Document document, Definitions bpmnModel, String type, String nsName) {
		String interfaceName = null;
		if (BPMNEditorUtils.isWSDL11File(type)) {
			interfaceName = "portType";
		} else if (BPMNEditorUtils.isWSDL20File(type)) {
			interfaceName = "interface";
		}
		NodeList ifNodes = document.getDocumentElement().getChildNodes();
		for (int i = 0; i < ifNodes.getLength(); i++) {
			Node ifNode = ifNodes.item(i);
			if (ifNode.getLocalName() != null && ifNode.getLocalName().equals(interfaceName)) {
				Interface _interface = (Interface) bpmnModel.generateRootElement(RootElements.TYPE_INTERFACE);
				_interface.setName(((Element) ifNode).getAttribute("name"));
				_interface.setImplementationRef(nsName + ":" + _interface.getName());
				bpmnModel.addRootElement(_interface);
				NodeList opNodes = ifNode.getChildNodes();
				for (int j = 0; j < opNodes.getLength(); j++) {
					Node opNode = opNodes.item(j);
					if (opNode.getLocalName() != null && opNode.getLocalName().equals("operation")) {
						Operation op = (Operation) _interface.getOperations().generateNewElement();
						op.setName(((Element) opNode).getAttribute("name"));
						op.setImplementationRef(nsName + ":" + op.getName());
						_interface.getOperations().add(op);
						NodeList ioNodes = opNode.getChildNodes();
						for (int k = 0; k < ioNodes.getLength(); k++) {
							Node ioNode = ioNodes.item(k);
							String nodeName = ioNode.getLocalName();
							if (nodeName != null) {
								String msg = ((Element) ioNode).getAttribute("message");
								if (msg.length() == 0) {
									msg = ((Element) ioNode).getAttribute("element");
									if (msg.length() == 0) {
										String ref = ((Element) ioNode).getAttribute("ref");
										if (ref.length() == 0) {
											continue;
										} else {
											int index = ref.indexOf(":");
											if (index != -1) {
												ref = ref.substring(index + 1);
											}
											NodeList faultNodes = ifNode.getChildNodes();
											for (int l = 0; l < faultNodes.getLength(); l++) {
												Node fNode = faultNodes.item(l);
												if (fNode.getLocalName() != null && fNode.getLocalName().equals("fault")) {
													if (ref.equals(((Element) fNode).getAttribute("name"))) {
														msg = ((Element) fNode).getAttribute("element");
														break;
													}
												}
											}
										}
									}
								}

								if (msg.length() != 0) {
									int index = msg.indexOf(":");
									if (index != -1) {
										msg = msg.substring(index + 1);
									}
									msg = nsName + ":" + msg;
									Message message = BPMNModelUtils.getMessageByStructure(msg, bpmnModel);
									if (nodeName.equals("input")) {
										op.setInMessageRef(message.getId());
									} else if (nodeName.equals("output")) {
										op.setOutMessageRef(message.getId());
									} else {
										BPMNError error = BPMNModelUtils.getErrorByStructure(msg, bpmnModel);
										if (error == null) {
											error = (BPMNError) bpmnModel.generateRootElement(RootElements.TYPE_ERROR);
											error.setName("Error " + error.getId().substring(4));
											error.setStructureRef(BPMNModelUtils.getItemDefinitionByStructure(msg, bpmnModel).getId());
											bpmnModel.addRootElement(error);
										}
										op.addErrorRef(error.getId());
									}
								}
							}
						}
					}
				}
			}
		}
	}

}
