package org.yaoqiang.bpmn.editor.view;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.swing.JOptionPane;

import org.yaoqiang.bpmn.editor.BPMNEditor;
import org.yaoqiang.bpmn.editor.swing.BPMNGraphComponent;
import org.yaoqiang.bpmn.editor.util.BPMNEditorUtils;
import org.yaoqiang.bpmn.model.BPMNModelUtils;
import org.yaoqiang.bpmn.model.elements.XMLCollection;
import org.yaoqiang.bpmn.model.elements.XMLComplexElement;
import org.yaoqiang.bpmn.model.elements.XMLElement;
import org.yaoqiang.bpmn.model.elements.XMLTextElements;
import org.yaoqiang.bpmn.model.elements.activities.Activity;
import org.yaoqiang.bpmn.model.elements.activities.CallActivity;
import org.yaoqiang.bpmn.model.elements.activities.LoopCharacteristics;
import org.yaoqiang.bpmn.model.elements.activities.LoopTypes;
import org.yaoqiang.bpmn.model.elements.activities.MultiInstanceLoopCharacteristics;
import org.yaoqiang.bpmn.model.elements.activities.ReceiveTask;
import org.yaoqiang.bpmn.model.elements.activities.ResourceRoles;
import org.yaoqiang.bpmn.model.elements.activities.SubProcess;
import org.yaoqiang.bpmn.model.elements.artifacts.Artifact;
import org.yaoqiang.bpmn.model.elements.artifacts.Artifacts;
import org.yaoqiang.bpmn.model.elements.artifacts.Association;
import org.yaoqiang.bpmn.model.elements.artifacts.Category;
import org.yaoqiang.bpmn.model.elements.artifacts.CategoryValue;
import org.yaoqiang.bpmn.model.elements.artifacts.Group;
import org.yaoqiang.bpmn.model.elements.artifacts.TextAnnotation;
import org.yaoqiang.bpmn.model.elements.choreography.Choreography;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.ChoreographyActivity;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.ChoreographyActivity.ChoreographyLoopType;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.ChoreographyTask;
import org.yaoqiang.bpmn.model.elements.choreographyactivities.SubChoreography;
import org.yaoqiang.bpmn.model.elements.collaboration.Collaboration;
import org.yaoqiang.bpmn.model.elements.collaboration.MessageFlow;
import org.yaoqiang.bpmn.model.elements.collaboration.Participant;
import org.yaoqiang.bpmn.model.elements.conversations.ConversationLink;
import org.yaoqiang.bpmn.model.elements.conversations.ConversationNode;
import org.yaoqiang.bpmn.model.elements.conversations.ConversationNodes;
import org.yaoqiang.bpmn.model.elements.core.common.FlowElement;
import org.yaoqiang.bpmn.model.elements.core.common.FlowElements;
import org.yaoqiang.bpmn.model.elements.core.common.FlowElementsContainer;
import org.yaoqiang.bpmn.model.elements.core.common.FlowNode;
import org.yaoqiang.bpmn.model.elements.core.common.Message;
import org.yaoqiang.bpmn.model.elements.core.common.SequenceFlow;
import org.yaoqiang.bpmn.model.elements.core.foundation.BaseElement;
import org.yaoqiang.bpmn.model.elements.core.foundation.Documentations;
import org.yaoqiang.bpmn.model.elements.core.foundation.RootElements;
import org.yaoqiang.bpmn.model.elements.core.infrastructure.Definitions;
import org.yaoqiang.bpmn.model.elements.data.DataAssociation;
import org.yaoqiang.bpmn.model.elements.data.DataInput;
import org.yaoqiang.bpmn.model.elements.data.DataObject;
import org.yaoqiang.bpmn.model.elements.data.DataObjectReference;
import org.yaoqiang.bpmn.model.elements.data.DataOutput;
import org.yaoqiang.bpmn.model.elements.data.DataStore;
import org.yaoqiang.bpmn.model.elements.data.DataStoreReference;
import org.yaoqiang.bpmn.model.elements.data.InputOutputSpecification;
import org.yaoqiang.bpmn.model.elements.data.ItemAwareElement;
import org.yaoqiang.bpmn.model.elements.data.Properties;
import org.yaoqiang.bpmn.model.elements.events.BoundaryEvent;
import org.yaoqiang.bpmn.model.elements.events.CatchEvent;
import org.yaoqiang.bpmn.model.elements.events.Event;
import org.yaoqiang.bpmn.model.elements.events.EventDefinition;
import org.yaoqiang.bpmn.model.elements.events.EventDefinitions;
import org.yaoqiang.bpmn.model.elements.events.LinkEventDefinition;
import org.yaoqiang.bpmn.model.elements.events.ThrowEvent;
import org.yaoqiang.bpmn.model.elements.gateways.EventBasedGateway;
import org.yaoqiang.bpmn.model.elements.process.BPMNProcess;
import org.yaoqiang.bpmn.model.elements.process.GlobalTask;
import org.yaoqiang.bpmn.model.elements.process.Lane;
import org.yaoqiang.graph.editor.swing.GraphComponent;
import org.yaoqiang.graph.model.YGraphModel;
import org.yaoqiang.graph.util.Constants;
import org.yaoqiang.graph.util.Utils;

import com.mxgraph.model.mxCell;
import com.mxgraph.model.mxGeometry;
import com.mxgraph.model.mxGraphModel;
import com.mxgraph.util.mxEvent;
import com.mxgraph.util.mxEventObject;
import com.mxgraph.util.mxEventSource;
import com.mxgraph.util.mxResources;
import com.mxgraph.util.mxUndoableEdit;
import com.mxgraph.util.mxUndoableEdit.mxUndoableChange;

/**
 * BPMNModelManager
 * 
 * @author Shi Yaoqiang(shi_yaoqiang@yahoo.com)
 */
public class BPMNModelManager extends mxEventSource {

	protected BPMNEditor editor;

	protected BPMNGraphComponent graphComponent;

	protected BPMNGraph graph;

	protected YGraphModel model;

	protected Map<String, XMLElement> bpmnElementMap;

	protected Definitions bpmnModel;

	public BPMNModelManager(BPMNEditor editor) {
		if (this.graph != null) {
			this.graph.removeListener(handler);
		}

		this.editor = editor;
		this.graphComponent = (BPMNGraphComponent) editor.getCurrentGraphComponent();
		this.graph = graphComponent.getGraph();
		this.model = graph.getModel();
		this.bpmnModel = graph.getBpmnModel();
		this.bpmnElementMap = graph.getBpmnElementMap();
		graph.setBpmnModelManager(this);

		if (this.graph != null) {
			this.graph.addListener(mxEvent.CELLS_ADDED, handler);
			this.graph.addListener(mxEvent.CELLS_REMOVED, handler);
			this.graph.addListener(mxEvent.CHANGE, handler);
			this.graph.addListener(mxEvent.LABEL_CHANGED, handler);
			this.graph.addListener(Constants.EVENT_LOOPTYPE_CHANGED, handler);
			this.graph.addListener(Constants.EVENT_ATTR_CHANGED, handler);
		}
	}

	protected mxIEventListener handler = new mxIEventListener() {
		public void invoke(Object source, mxEventObject evt) {
			if (BPMNEditor.isNativeFormat()) {
				return;
			}
			String eventName = evt.getName();
			Map<String, Object> properties = evt.getProperties();
			Object[] cells = (Object[]) properties.get("cells");
			try {
				List<Object> last = new ArrayList<Object>();
				if (eventName.equals(mxEvent.CELLS_ADDED)) {
					if (cells != null && cells.length > 0) {
						List<Object> addedCells = new ArrayList<Object>();
						for (Object cell : cells) {
							if (((mxCell) cell).isEdge()) {
								addedCells.add(cell);
							} else if (graph.isChoreographyParticipant(cell) || graph.isChoreographyTask(cell) || graph.isChoreographySubprocess(cell)) {
								if (cells.length == 3) {
									continue;
								} else {
									addElement((mxCell) cell);
								}
							} else if (graph.isChoreography(cell) || graph.isSubChoreography(cell)) {
								if (model.getChildCount(cell) > 0) {
									removeCell((mxCell) cell);
									for (Object child : graph.getChildVertices(cell)) {
										if (graph.isChoreographyParticipant(child)) {
											addedCells.add(child);
										} else {
											addElement((mxCell) child);
										}
									}
								} else {
									continue;
								}
							} else {
								mxCell previous = (mxCell) properties.get("previous");
								if (previous != null && previous != model.getParent(cell)) {
									removeFromPreviousParent(previous, (mxCell) cell);
								}
								changeCell((mxCell) cell);
							}
						}
						for (Object cell : addedCells) {
							changeCell((mxCell) cell);
						}
					}
				} else if (eventName.equals(mxEvent.CHANGE)) {
					if (cells != null && cells.length > 0) {
						for (Object cell : cells) {
							changeCell((mxCell) cell);
						}
					}
				} else if (eventName.equals(mxEvent.CELLS_REMOVED)) {
					if (cells != null && cells.length > 0) {
						List<Object> removedCells = new ArrayList<Object>();
						for (Object cell : cells) {
							if (!((mxCell) cell).isEdge()) {
								if (graph.isCallActivityProcess(cell)) {
									GraphComponent graphComponent = editor.getGraphComponent(((mxCell) cell).getId());
									editor.getCenterPane().remove(graphComponent);
								}
								removedCells.add(cell);
							} else {
								removeCell((mxCell) cell);
							}
						}
						for (Object cell : removedCells) {
							removeCell((mxCell) cell);
						}
					}
				} else if (eventName.equals(Constants.EVENT_LOOPTYPE_CHANGED)) {
					if (cells != null && cells.length == 1) {
						changeCellLoopType((mxCell) cells[0]);
					}
				} else if (eventName.equals(Constants.EVENT_ATTR_CHANGED)) {
					if (cells != null && cells.length > 0) {
						for (Object cell : cells) {
							changeCellAttr((mxCell) cell, null);
						}
					}
				} else if (eventName.equals(mxEvent.LABEL_CHANGED)) {
					if (cells != null && cells.length == 1) {
						mxCell cell = (mxCell) cells[0];
						if (graph.isCallActivityProcess(cell)) {
							String calledElement = ((CallActivity) graph.getBpmnElementMap().get(cell.getId())).getCalledElement();
							mxCell diagramCell = (mxCell) graph.getModel().getCell(calledElement);
							if (diagramCell != null) {
								diagramCell.setValue((String) cell.getValue());
								BPMNEditorUtils.refreshDiagramList(editor, graph.getCurrentRoot());
							}
							if (editor.getGraphComponents().get(calledElement) != null) {
								editor.getCenterPane().setTabTitle(calledElement, (String) cell.getValue());
							}
						}
						changeCellLabel((mxCell) cells[0], null);
					}
				} else if (eventName.equals(mxEvent.UNDO) || eventName.equals(mxEvent.REDO)) {
					mxUndoableEdit edit = (mxUndoableEdit) evt.getProperty("edit");
					boolean isAttachedEvent = false;
					boolean remove = false;
					Set<Object> parentAdded = new HashSet<Object>();
					Set<Object> childRemoved = new HashSet<Object>();
					for (mxUndoableChange change : edit.getChanges()) {
						if (change instanceof mxGraphModel.mxChildChange) {
							int index = ((mxGraphModel.mxChildChange) change).getIndex();
							int previousIndex = ((mxGraphModel.mxChildChange) change).getPreviousIndex();
							Object parent = ((mxGraphModel.mxChildChange) change).getParent();
							Object previous = ((mxGraphModel.mxChildChange) change).getPrevious();
							mxCell child = (mxCell) ((mxGraphModel.mxChildChange) change).getChild();
							if (previousIndex == -1 && previousIndex != index) {
								if (parentAdded.contains(parent)) {
									continue;
								}
								if (graph.isChoreographyParticipant(child)) {
									last.add(child);
								} else if (graph.isChoreography(child) || graph.isSubChoreography(child)) {
									if (edit.getChanges().size() > 1) {
										continue;
									} else {
										Object[] children = graph.getChildCells(child);
										if (children != null && children.length > 0) {
											for (Object cell : children) {
												if (graph.isChoreographyParticipant(cell)) {
													last.add(cell);
												} else {
													addElement((mxCell) cell);
													parentAdded.add(cell);
												}
											}
										}
									}
								} else {
									addElement(child);
								}
								parentAdded.add(child);
							} else if (index == 1 || previousIndex == 1 || index == 0 || (previousIndex == 0 || index == -1) && previousIndex != index) {
								if (parent == null || parent == previous && (index != 1 && previousIndex != 1 || index == -1 || previousIndex == 1)) {
									if (parentAdded.contains(child) || childRemoved.contains(child)) {
										continue;
									}
									removeCell(child);
									remove = true;

									if ((graph.isChoreography(child) || graph.isSubChoreography(child))) {
										childRemoved.addAll(Arrays.asList(graph.getChildVertices(child)));
									} else {
										childRemoved.add(child);
									}
								} else {
									if (parentAdded.contains(child) || childRemoved.contains(child)) {
										continue;
									}
									changeCell(child);
									if ((graph.isChoreography(child) || graph.isSubChoreography(child))) {
										parentAdded.addAll(Arrays.asList(graph.getChildVertices(child)));
									} else {
										parentAdded.add(child);
									}
								}
							}
							if (graph.isLinkEvent(child)) {
								resetLinkEvent(child);
							}
							if (isAttachedEvent) {
								editor.getBpmnView().refreshView(graph);
								return;
							}
						} else if (change instanceof mxGraphModel.mxValueChange) {
							mxCell cell = (mxCell) ((mxGraphModel.mxValueChange) change).getCell();
							if (graph.isCallActivityProcess(cell) && editor.getGraphComponents().get(cell.getId()) != null) {
								editor.getCenterPane().setTabTitle(cell.getId(), (String) cell.getValue());
							}
							changeCellLabel(cell, ((mxGraphModel.mxValueChange) change).getPrevious().toString());
						} else if (change instanceof mxGraphModel.mxGeometryChange) {
							mxCell event = (mxCell) ((mxGraphModel.mxGeometryChange) change).getCell();
							if (graph.isAttachedEvent(event) || graph.isAttachedMessage(event)) {
								isAttachedEvent = true;
							}
						} else if (change instanceof mxGraphModel.mxStyleChange) {
							if (((mxGraphModel.mxStyleChange) change).getStyle().length() == 0 || remove) {
								editor.getBpmnView().refreshView(graph);
								return;
							}
							mxCell cell = (mxCell) ((mxGraphModel.mxStyleChange) change).getCell();
							String oldStyle = ((mxGraphModel.mxStyleChange) change).getPrevious();
							if (graph.isChoreographyParticipant(cell) || graph.isPool(cell) || graph.isDataObject(cell) || graph.isEvent(cell)
									|| graph.isTask(cell) || graph.isSequenceFlow(cell)) {
								if (edit.getChanges().size() == 1 && !graph.isEvent(cell)) {
									changeCellAttr(cell, oldStyle);
								} else {
									changeCell(cell);
								}
								if (graph.isLinkEvent(cell)) {
									resetLinkEvent(cell);
								}
							} else {
								editor.getBpmnView().refreshView(graph);
								return;
							}
						}
					}
					for (Object cell : last) {
						addElement((mxCell) cell);
						parentAdded.add(cell);
					}
					editor.getBpmnView().refreshView(graph);
				}
				if (cells != null && cells.length > 0) {
					for (Object cell : cells) {
						if (graph.isLinkEvent(cell)) {
							List<Object> events = new ArrayList<Object>();
							Event e = (Event) bpmnElementMap.get(((mxCell) cell).getId());
							if (e == null) {
								events.addAll(Arrays.asList(graph.getLinkEvent(model.getParent(cell), !graph.isThrowEvent(cell), ((mxCell) cell).getValue()
										.toString())));
							} else {
								events.add(cell);
							}
							for (Object el : events) {
								resetLinkEvent((mxCell) el);
							}
						}
					}
				}
			} catch (Exception e) {
				e.printStackTrace();
				JOptionPane
						.showMessageDialog(null, e.getStackTrace(), "Please Capture This Error Screen Shots and Submit this BUG.", JOptionPane.ERROR_MESSAGE);
			}
		}
	};

	private void addElement(mxCell cell) {
		FlowElementsContainer container = getFlowElementsContainer(cell);

		if (graph.isSwimlane(cell)) {
			addSwimlaneElement(cell);
		} else if (graph.isMessage(cell)) {
			addMessageElement(cell);
		} else if (graph.isChoreographyParticipant(cell)) {
			addChoreographyParticipantElement(container, cell);
		} else if (graph.isConversation(cell)) {
			addConversationNodeElement(cell);
		} else if (graph.isDataInput(cell) || graph.isDataOutput(cell)) {
			addDataElement(container, cell);
		} else if (graph.isArtifact(cell)) {
			addArtifactElement(cell);
		} else if (model.isEdge(cell)) {
			if (cell.getSource() == null || cell.getTarget() == null) {
				model.remove(cell);
			} else {
				String style = graph.adjustEdgeStyle(cell);
				if (style.equals(Constants.EDGE_TYPE_SEQUENCE_FLOW)) {
					addSequenceFlowElement(container, cell);
				} else if (style.equals(Constants.EDGE_TYPE_ASSOCIATION) || style.equals(Constants.EDGE_TYPE_COMPENSATION_ASSOCIATION)) {
					addArtifactElement(cell);
				} else if (style.equals(Constants.EDGE_TYPE_DATA_ASSOCIATION)) {
					addDataAssociationElement(cell);
				} else if (style.equals(Constants.EDGE_TYPE_MESSAGE_FLOW)) {
					addMessageFlowElement(cell);
				} else if (style.equals(Constants.EDGE_TYPE_CONVERSATION_LINK)) {
					addConversationLinkElement(cell);
				}
				model.getCells().put(cell.getId(), cell);
			}
		} else if (container != null) {
			FlowElement flowElement = (FlowElement) container.generateFlowElement(getFlowElementType(cell));
			if (flowElement != null) {
				flowElement.setId(cell.getId());
				if (flowElement instanceof DataStoreReference) {

				} else {
					flowElement.setName(cell.getValue().toString());
				}

				if (flowElement instanceof Event) {
					if (flowElement instanceof BoundaryEvent) {
						((BoundaryEvent) flowElement).setAttachedToRef(cell.getParent().getId());
						if (!graph.cancelActivity(cell)) {
							((BoundaryEvent) flowElement).setCancelActivity(false);
						}
					}
					if (flowElement instanceof CatchEvent) {
						if (graph.isParallelMultipleEvent(cell)) {
							((CatchEvent) flowElement).setParallelMultiple(true);
						}
					}

					EventDefinition eventDefinition = ((Event) flowElement).generateEventDefinition(getEventDefinitionType(cell));
					((Event) flowElement).addEventDefinition(eventDefinition);
					if (eventDefinition instanceof LinkEventDefinition) {
						((LinkEventDefinition) eventDefinition).setName(flowElement.getName());
					}
				} else if (flowElement instanceof DataObjectReference) {
					String value = cell.getValue().toString();
					if (value.length() != 0) {
						int index = value.lastIndexOf("\n[");
						if (index > 0) {
							value = value.substring(0, index);
						}
					}
					boolean isCollection = graph.isCollectionDataObject(cell);
					DataObject dataObject = null;
					List<XMLElement> dataObjects = flowElement.getParent().getDataObjects(isCollection);
					if (dataObjects.isEmpty()) {
						dataObject = (DataObject) container.generateFlowElement(FlowElements.TYPE_DATAOBJECT);
						dataObject.setName(mxResources.get("dataObject"));
						dataObject.setIsCollection(isCollection);
						container.addFlowElement(dataObject);
						bpmnElementMap.put(dataObject.getId(), dataObject);
					} else {
						for (XMLElement d : dataObjects) {
							dataObject = (DataObject) d;
							if (dataObject.getName().equals(value)) {
								break;
							}
						}
					}
					cell.setValue(dataObject.getName());
					((DataObjectReference) flowElement).setDataObjectRef(dataObject.getId());
				} else if (flowElement instanceof DataStoreReference) {
					DataStore dataStore = null;
					if (bpmnModel.getDataStores().isEmpty()) {
						dataStore = (DataStore) bpmnModel.generateRootElement(RootElements.TYPE_DATASTORE);
						dataStore.setName(mxResources.get("dataStore"));
						bpmnModel.addRootElement(dataStore);
						bpmnElementMap.put(dataStore.getId(), dataStore);
					} else {
						for (XMLElement ds : bpmnModel.getDataStores()) {
							dataStore = (DataStore) ds;
							if (dataStore.getName().equals(cell.getValue())) {
								break;
							}

						}
					}
					cell.setValue(dataStore.getName());
					((DataStoreReference) flowElement).setDataStoreRef(dataStore.getId());
				} else if (flowElement instanceof ChoreographyActivity) {

				} else if (flowElement instanceof SubProcess) {
					if (graph.isEventSubProcess(cell)) {
						((SubProcess) flowElement).setTriggeredByEvent(true);
					}
				} else if (flowElement instanceof EventBasedGateway) {
					if (graph.isInstantiateEventGateway(cell)) {
						EventBasedGateway eventGateway = (EventBasedGateway) flowElement;
						eventGateway.setInstantiate(true);
						if (graph.isParallelEventGateway(cell)) {
							eventGateway.setEventGatewayType(EventBasedGateway.EventGatewayType.Parallel.toString());
						}
					}
				} else if (flowElement instanceof CallActivity) {
					if (graph.isCallActivity(cell)) {
						GlobalTask globalTask = null;
						if (graph.isAbstractTask(cell)) {
							globalTask = (GlobalTask) bpmnModel.generateRootElement(RootElements.TYPE_GLOBAL_TASK);
						} else if (graph.isUserTask(cell)) {
							globalTask = (GlobalTask) bpmnModel.generateRootElement(RootElements.TYPE_GLOBAL_USER_TASK);
						} else if (graph.isScriptTask(cell)) {
							globalTask = (GlobalTask) bpmnModel.generateRootElement(RootElements.TYPE_GLOBAL_SCRIPT_TASK);
						} else if (graph.isManualTask(cell)) {
							globalTask = (GlobalTask) bpmnModel.generateRootElement(RootElements.TYPE_GLOBAL_MANUAL_TASK);
						} else if (graph.isBusinessRuleTask(cell)) {
							globalTask = (GlobalTask) bpmnModel.generateRootElement(RootElements.TYPE_GLOBAL_BUSINESS_RULE_TASK);
						}
						globalTask.setId("GT" + cell.getId());
						globalTask.setName("Global Task");
						bpmnModel.addRootElement(globalTask);
						bpmnElementMap.put(globalTask.getId(), globalTask);
						((CallActivity) flowElement).setCalledElement(globalTask.getId());
					} else if (graph.isCallActivityProcess(cell)) {
						String id = "P" + cell.getId();
						BPMNProcess newProcess = (BPMNProcess) bpmnElementMap.get(id);
						if (newProcess == null) {
							newProcess = (BPMNProcess) bpmnModel.generateRootElement(RootElements.TYPE_PROCESS);
						}
						newProcess.setId(id);
						newProcess.setName("Process");
						bpmnModel.addRootElement(newProcess);
						bpmnElementMap.put(newProcess.getId(), newProcess);
						((CallActivity) flowElement).setCalledElement(newProcess.getId());
					}
				}

				String loopType = "";
				if (flowElement instanceof Activity) {
					if (graph.isCompensationActivity(cell)) {
						((Activity) flowElement).setIsForCompensation(true);
					}
					if (flowElement instanceof ReceiveTask) {
						ReceiveTask receiveTask = (ReceiveTask) flowElement;
						if (graph.isInstantiateReceiveTask(cell)) {
							receiveTask.setInstantiate(true);
						} else {
							receiveTask.setInstantiate(false);
						}
					}
				} else if (flowElement instanceof ChoreographyActivity) {
					if (graph.isMultiInstanceParallelActivity(cell)) {
						loopType = ChoreographyLoopType.MultiInstanceParallel.toString();
					} else if (graph.isMultiInstanceSequentialActivity(cell)) {
						loopType = ChoreographyLoopType.MultiInstanceSequential.toString();
					} else if (graph.isStandardLoopActivity(cell)) {
						loopType = ChoreographyLoopType.Standard.toString();
					} else {
						loopType = ChoreographyLoopType.None.toString();
					}
					((ChoreographyActivity) flowElement).setLoopType(loopType);
				}
				for (String categoryValueId : BPMNModelUtils.getCachedCategoryValueIds(bpmnModel, flowElement.getId())) {
					flowElement.addCategoryValueRef(categoryValueId);
				}
				container.addFlowElement(flowElement);
				bpmnElementMap.put(flowElement.getId(), flowElement);
			}
		}
		List<Object> edges = new ArrayList<Object>();
		Object[] cells = graph.getChildCells(cell);
		if (cells != null && cells.length > 0) {
			for (Object child : cells) {
				if (((mxCell) child).isEdge() || graph.isChoreographyParticipant(child)) {
					edges.add(child);
				} else {
					addElement((mxCell) child);
				}
			}
			for (Object child : edges) {
				addElement((mxCell) child);
			}
		}
	}

	private void addChoreographyParticipantElement(FlowElementsContainer container, mxCell cell) {
		int index = cell.getId().indexOf("_part_");
		String actId = cell.getId().substring(0, index);
		String partId = cell.getId().substring(index + 6);
		Choreography choreography = null;
		if (container instanceof SubChoreography) {
			choreography = (Choreography) ((SubChoreography) container).getParent().getParent();
		} else {
			choreography = (Choreography) container;
		}
		if (choreography == null) {
			return;
		}

		if (bpmnElementMap.get(partId) == null) {
			Participant participant = choreography.addParticipant(partId, cell.getValue().toString());
			bpmnElementMap.put(participant.getId(), participant);
		}
		ChoreographyActivity ca = (ChoreographyActivity) bpmnElementMap.get(actId);
		if (ca == null) {
			return;
		}
		if (!ca.getParticipantList().contains(partId)) {
			ca.addParticipantRef(partId);
		}

		if (graph.isInitiatingChoreographyParticipant(cell)) {
			ca.setInitiatingParticipantRef(partId);
		}

		if (ca instanceof ChoreographyTask) {
			addMessageFlowForChoreographyTask(choreography, (ChoreographyTask) ca);
		} else {
			if (graph.isAdditionalChoreographyParticipant(cell)) {
				mxGeometry subgeo = model.getGeometry(Utils.getChoreographyActivity(graph, cell.getParent()));
				double yOffset = 0;
				if (graph.isTopChoreographyParticipant(cell)) {
					yOffset = subgeo.getY() - 1;
				} else {
					yOffset = subgeo.getY() + subgeo.getHeight() - 1;
				}

				mxGeometry partgeo = model.getGeometry(cell);
				partgeo.setY(yOffset);
				model.setGeometry(cell, partgeo);

				Utils.arrangeChoreography(graph, cell.getParent(), false);

				graph.refresh();
			}
		}
	}

	private void addMessageFlowForChoreographyTask(Choreography choreography, ChoreographyTask ct) {
		String initPartId = ct.getInitiatingParticipantRef();
		ct.clearMessageFlowRefs();

		if (initPartId.length() != 0) {
			for (XMLElement participantRef : ct.getParticipantRefList()) {
				if (!initPartId.equals(participantRef.toValue())) {
					String id = choreography.getId() + initPartId + participantRef.toValue();
					choreography.addMessageFlow(id, initPartId, participantRef.toValue());
					ct.addMessageFlowRef(id);
					id = choreography.getId() + participantRef.toValue() + initPartId;
					choreography.addMessageFlow(id, participantRef.toValue(), initPartId);
					ct.addMessageFlowRef(id);
				}
			}
		}

		for (XMLElement mf : choreography.getMessageFlowList()) {
			if (!BPMNModelUtils.hasMessageFlowRef(choreography, ((MessageFlow) mf).getId())) {
				choreography.getMessageFlows().remove(((MessageFlow) mf).getId());
			}
		}
	}

	private void addConversationLinkElement(mxCell cell) {
		Collaboration collaboration = BPMNModelUtils.getCollaboration(bpmnModel);
		ConversationLink conversationLink = collaboration.addConversationLink(cell.getId(), cell.getSource().getId(), cell.getTarget().getId());
		bpmnElementMap.put(conversationLink.getId(), conversationLink);
	}

	private void addConversationNodeElement(mxCell cell) {
		Collaboration collaboration = BPMNModelUtils.getCollaboration(bpmnModel);
		if (collaboration == null) {
			collaboration = bpmnModel.createCollaboration();
			XMLElement rootElement = bpmnElementMap.get("_1");
			if (rootElement instanceof BPMNProcess) {
				bpmnElementMap.put("P_1", rootElement);
			} else if (rootElement instanceof Choreography) {
				bpmnElementMap.put("C_1", rootElement);
			}
			bpmnElementMap.put(collaboration.getId(), collaboration);
		}
		ConversationNode conversationNode = collaboration.addConversationNode(getConversationNodeType(cell), cell.getId(), cell.getValue().toString());
		bpmnElementMap.put(conversationNode.getId(), conversationNode);
	}

	private void addMessageFlowElement(mxCell cell) {
		Collaboration collaboration = BPMNModelUtils.getCollaboration(bpmnModel);
		MessageFlow messageFlow = collaboration.addMessageFlow(cell.getId(), cell.getSource().getId(), cell.getTarget().getId());
		bpmnElementMap.put(messageFlow.getId(), messageFlow);
	}

	private void addDataAssociationElement(mxCell cell) {
		ItemAwareElement data = null;
		String dataId = null;
		DataAssociation da = null;
		mxCell sCell = (mxCell) cell.getSource();
		mxCell tCell = (mxCell) cell.getTarget();
		int opt = JOptionPane.YES_OPTION;
		if (graph.isDataObject(sCell) || graph.isDataStore(sCell)) {
			dataId = "Din" + tCell.getId() + sCell.getId();
			XMLElement actOrEvent = bpmnElementMap.get(tCell.getId());
			if (actOrEvent instanceof Activity) {
				Activity act = (Activity) actOrEvent;
				int size = act.getIoSpecification().getDataInputs().size();
				if (size > 0) {
					opt = JOptionPane.showConfirmDialog(null, mxResources.get("InfoYesToCreateNewInputOutputOrNoToSelectExistedInputOutput"),
							mxResources.get("InfoCreateOrSelect"), JOptionPane.YES_NO_OPTION);
				}
				if (opt == JOptionPane.YES_OPTION) {
					data = act.getIoSpecification().addDataInput(dataId, null, false);
				} else {
					editor.getBpmnElementDialog().initDialog().editBPMNElement(actOrEvent, "selectDataInput");
					data = (ItemAwareElement) act.getIoSpecification().getDataInputList().get(size - 1);
				}
				da = act.addDataInputAssociation(cell.getId(), sCell.getId(), data.getId());
			} else if (actOrEvent instanceof ThrowEvent) {
				ThrowEvent evt = (ThrowEvent) actOrEvent;
				int size = evt.getDataInputs().size();
				if (size > 0) {
					opt = JOptionPane.showConfirmDialog(null, mxResources.get("InfoYesToCreateNewInputOutputOrNoToSelectExistedInputOutput"),
							mxResources.get("InfoCreateOrSelect"), JOptionPane.YES_NO_OPTION);
				}
				if (opt == JOptionPane.YES_OPTION) {
					data = evt.addDataInput(dataId);
				} else {
					editor.getBpmnElementDialog().initDialog().editBPMNElement(actOrEvent, "selectDataInput");
					data = (ItemAwareElement) evt.getDataInputList().get(size - 1);
				}
				da = evt.addDataInputAssociation(cell.getId(), sCell.getId(), data.getId());
			}
		} else if (graph.isDataObject(tCell) || graph.isDataStore(tCell)) {
			dataId = "Dout" + sCell.getId() + tCell.getId();
			XMLElement actOrEvent = bpmnElementMap.get(sCell.getId());
			if (actOrEvent instanceof Activity) {
				Activity act = (Activity) actOrEvent;
				int size = act.getIoSpecification().getDataOutputs().size();
				if (size > 0) {
					opt = JOptionPane.showConfirmDialog(null, mxResources.get("InfoYesToCreateNewInputOutputOrNoToSelectExistedInputOutput"),
							mxResources.get("InfoCreateOrSelect"), JOptionPane.YES_NO_OPTION);
				}
				if (opt == JOptionPane.YES_OPTION) {
					data = act.getIoSpecification().addDataOutput(dataId, null, false);
				} else {
					editor.getBpmnElementDialog().initDialog().editBPMNElement(actOrEvent, "selectDataOutput");
					data = (ItemAwareElement) act.getIoSpecification().getDataOutputList().get(size - 1);
				}
				da = act.addDataOutputAssociation(cell.getId(), data.getId(), tCell.getId());
			} else if (actOrEvent instanceof CatchEvent) {
				CatchEvent evt = (CatchEvent) actOrEvent;
				int size = evt.getDataOutputs().size();
				if (size > 0) {
					opt = JOptionPane.showConfirmDialog(null, mxResources.get("InfoYesToCreateNewInputOutputOrNoToSelectExistedInputOutput"),
							mxResources.get("InfoCreateOrSelect"), JOptionPane.YES_NO_OPTION);
				}
				if (opt == JOptionPane.YES_OPTION) {
					data = evt.addDataOutput(dataId);
				} else {
					editor.getBpmnElementDialog().initDialog().editBPMNElement(actOrEvent, "selectDataOutput");
					data = (ItemAwareElement) evt.getDataOutputList().get(size - 1);
				}
				da = evt.addDataOutputAssociation(cell.getId(), data.getId(), tCell.getId());
			}
		}
		if (data != null && da != null) {
			bpmnElementMap.put(data.getId(), (XMLElement) data);
			bpmnElementMap.put(da.getId(), da);
		}
	}

	private void addSequenceFlowElement(FlowElementsContainer container, mxCell cell) {
		if (container == null) {
			return;
		}
		SequenceFlow sequenceFlow = (SequenceFlow) container.generateFlowElement(FlowElements.TYPE_SEQUENCE_FLOW);
		String sourceId = cell.getSource().getId();
		String targetId = cell.getTarget().getId();
		sequenceFlow.setId(cell.getId());
		sequenceFlow.setName(cell.getValue().toString());
		sequenceFlow.setSourceRef(sourceId);
		sequenceFlow.setTargetRef(targetId);
		FlowNode srcNode = (FlowNode) container.getFlowElements().get(sourceId);
		if (srcNode == null) {
			srcNode = (FlowNode) bpmnElementMap.get(sourceId);
		}
		FlowNode tgtNode = (FlowNode) container.getFlowElements().get(targetId);
		if (tgtNode == null) {
			tgtNode = (FlowNode) bpmnElementMap.get(targetId);
		}
		if (srcNode == null || tgtNode == null) {
			return;
		}
		srcNode.addOutgoing(sequenceFlow.getId());
		tgtNode.addIncoming(sequenceFlow.getId());
		if (graph.isDefaultSequenceFlow(cell)) {
			srcNode.set("default", sequenceFlow.getId());
		} else {
			sequenceFlow.setConditionExpression(model.getAttrFromStyle("expression", cell.getStyle()));
		}
		for (String categoryValueId : BPMNModelUtils.getCachedCategoryValueIds(bpmnModel, sequenceFlow.getId())) {
			sequenceFlow.addCategoryValueRef(categoryValueId);
		}
		container.addFlowElement(sequenceFlow);
		bpmnElementMap.put(sequenceFlow.getId(), sequenceFlow);
	}

	private void addArtifactElement(mxCell cell) {
		Artifact artifact = null;
		Artifacts artifacts = null;

		Collaboration collaboration = BPMNModelUtils.getCollaboration(bpmnModel);
		Choreography choreography = BPMNModelUtils.getChoreography(bpmnModel);
		BPMNProcess process = BPMNModelUtils.getDefaultProcess(bpmnModel);
		if (collaboration != null) {
			artifacts = collaboration.getArtifacts();
		} else if (choreography != null) {
			artifacts = choreography.getArtifacts();
		} else if (process != null) {
			artifacts = process.getArtifacts();
		} else {
			process = bpmnModel.createProcess(true);
			bpmnElementMap.put(process.getId(), process);
			artifacts = process.getArtifacts();
		}

		if (graph.isAnnotation(cell)) {
			artifacts.setType(Artifacts.TYPE_TEXTANNOTATION);
			artifact = (Artifact) artifacts.generateNewElement();
			artifact.setId(cell.getId());
			((TextAnnotation) artifact).setText(cell.getValue().toString());
		} else if (graph.isGroupArtifact(cell)) {
			artifacts.setType(Artifacts.TYPE_GROUP);
			artifact = (Artifact) artifacts.generateNewElement();
			artifact.setId(cell.getId());
			CategoryValue categoryValue = bpmnModel.getDefaultCategoryValue();
			Category category = null;
			String groupName = "";
			if (categoryValue == null) {
				if (bpmnModel.getCategories().isEmpty()) {
					category = (Category) bpmnModel.generateRootElement(RootElements.TYPE_CATEGORY);
					bpmnModel.addRootElement(category);
					bpmnElementMap.put(category.getId(), category);
				} else {
					category = (Category) bpmnModel.getCategories().get(0);
				}
				categoryValue = category.addCategoryValue(mxResources.get("group"));
				bpmnElementMap.put(categoryValue.getId(), categoryValue);
			} else {
				category = ((Category) categoryValue.getParent().getParent());
			}
			((Group) artifact).setCategoryValueRef(categoryValue.getId());
			groupName = (category.getName().length() == 0 ? "" : category.getName() + ":") + categoryValue.getValue();
			cell.setValue(groupName);
		} else {
			artifacts.setType(Artifacts.TYPE_ASSOCIATION);
			artifact = (Artifact) artifacts.generateNewElement();
			artifact.setId(cell.getId());
			((Association) artifact).setSourceRef(cell.getSource().getId());
			((Association) artifact).setTargetRef(cell.getTarget().getId());
		}
		artifacts.add(artifact);
		bpmnElementMap.put(artifact.getId(), artifact);
	}

	private void addSwimlaneElement(mxCell cell) {
		Lane lane = null;
		mxCell parent = (mxCell) cell.getParent();
		if (graph.isPool(parent)) {
			BPMNProcess process = getProcessByParticipant(parent.getId());
			lane = process.addLane(cell.getId(), cell.getValue().toString());
			bpmnElementMap.put(lane.getId(), lane);
		} else if (graph.isLane(parent)) {
			Lane parentLane = (Lane) bpmnElementMap.get(parent.getId());
			lane = parentLane.getChildLaneSet().addLane(cell.getId(), cell.getValue().toString());
			bpmnElementMap.put(lane.getId(), lane);
		} else if (graph.isPlane(parent)) {
			Participant participant = bpmnModel.addParticipant(cell.getId(), cell.getValue().toString());
			bpmnElementMap.put(participant.getId(), participant);
			XMLElement rootElement = bpmnElementMap.get("_1");
			if (rootElement instanceof BPMNProcess) {
				bpmnElementMap.put("P_1", rootElement);
			} else if (rootElement instanceof Choreography) {
				bpmnElementMap.put("C_1", rootElement);
			}
			bpmnElementMap.put("_1", participant.getParent().getParent());
		}
	}

	private void addMessageElement(mxCell cell) {
		Message message = (Message) bpmnModel.generateRootElement(RootElements.TYPE_MESSAGE);
		message.setId(cell.getId());
		bpmnModel.addRootElement(message);
		bpmnElementMap.put(message.getId(), message);
		if (graph.isAttachedMessage(cell)) {
			MessageFlow messageFlow = (MessageFlow) bpmnElementMap.get(cell.getParent().getId());
			messageFlow.setMessageRef(message.getId());
		}
	}

	private void addDataElement(FlowElementsContainer container, mxCell cell) {
		InputOutputSpecification ioSpec = null;
		if (container instanceof BPMNProcess) {
			ioSpec = ((BPMNProcess) container).getIoSpecification();
		} else if (container instanceof SubProcess) {
			ioSpec = ((SubProcess) container).getIoSpecification();
		} else if (container == null) {
			return;
		}
		ItemAwareElement data = null;
		if (graph.isDataInput(cell)) {
			data = ioSpec.addDataInput(cell.getId(), cell.getValue().toString(), graph.isCollectionDataObject(cell));
		} else {
			data = ioSpec.addDataOutput(cell.getId(), cell.getValue().toString(), graph.isCollectionDataObject(cell));
		}
		bpmnElementMap.put(data.getId(), (XMLElement) data);
	}

	private void changeCellLoopType(mxCell cell) {
		XMLElement bpmnElement = bpmnElementMap.get(cell.getId());
		String loopType = "";
		if (bpmnElement instanceof Activity) {
			if (graph.isMultiInstanceParallelActivity(cell)) {
				loopType = LoopTypes.TYPE_MULTI_INSTANCE_LOOPCHARACTERISTICS;
			} else if (graph.isMultiInstanceSequentialActivity(cell)) {
				loopType = LoopTypes.TYPE_MULTI_INSTANCE_LOOPCHARACTERISTICS;
			} else if (graph.isStandardLoopActivity(cell)) {
				loopType = LoopTypes.TYPE_STANDARD_LOOPCHARACTERISTICS;
			}
			if (loopType.length() != 0) {
				LoopCharacteristics loopCharacteristics = ((Activity) bpmnElement).setLoopCharacteristics(loopType);
				if (loopCharacteristics instanceof MultiInstanceLoopCharacteristics) {
					if (graph.isMultiInstanceSequentialActivity(cell)) {
						((MultiInstanceLoopCharacteristics) loopCharacteristics).setIsSequential(true);
					}
				}
			}
		} else if (bpmnElement instanceof ChoreographyActivity) {
			if (graph.isMultiInstanceParallelActivity(cell)) {
				loopType = ChoreographyLoopType.MultiInstanceParallel.toString();
			} else if (graph.isMultiInstanceSequentialActivity(cell)) {
				loopType = ChoreographyLoopType.MultiInstanceSequential.toString();
			} else if (graph.isStandardLoopActivity(cell)) {
				loopType = ChoreographyLoopType.Standard.toString();
			} else {
				loopType = ChoreographyLoopType.None.toString();
			}
			((ChoreographyActivity) bpmnElement).setLoopType(loopType);
		}
	}

	private void changeCellAttr(mxCell cell, String previous) {
		String id = cell.getId();
		if (graph.isChoreographyParticipant(cell)) {
			id = id.substring(id.indexOf("_part_") + 6);
		}
		XMLElement bpmnElement = bpmnElementMap.get(id);
		if (bpmnElement instanceof Activity) {
			if (graph.isCompensationActivity(cell)) {
				((Activity) bpmnElement).setIsForCompensation(true);
			} else {
				((Activity) bpmnElement).setIsForCompensation(false);
			}
			if (graph.isReceiveTask(cell)) {
				ReceiveTask receiveTask = (ReceiveTask) bpmnElement;
				if (graph.isInstantiateReceiveTask(cell)) {
					receiveTask.setInstantiate(true);
				} else {
					receiveTask.setInstantiate(false);
				}
			}
		} else if (bpmnElement instanceof ItemAwareElement) {
			if (graph.isCollectionDataObject(cell)) {
				if (bpmnElement instanceof DataObjectReference) {
					DataObjectReference doRef = (DataObjectReference) bpmnElement;
					DataObject dataObject = doRef.getRefDataObject();
					if (dataObject != null) {
						dataObject.setIsCollection(true);
					}
				} else {
					((XMLComplexElement) bpmnElement).set("isCollection", "true");
				}
			} else if (graph.isDataObject(cell)) {
				if (bpmnElement instanceof DataObjectReference) {
					DataObjectReference doRef = (DataObjectReference) bpmnElement;
					DataObject dataObject = doRef.getRefDataObject();
					if (dataObject != null) {
						dataObject.setIsCollection(false);
					}
				} else {
					((XMLComplexElement) bpmnElement).set("isCollection", "false");
				}
			}
			if (bpmnElement instanceof DataObjectReference) {
				DataObjectReference doRef = (DataObjectReference) bpmnElement;
				DataObject dataObject = doRef.getRefDataObject();
				if (dataObject != null) {
					String dataState = ((DataObjectReference) bpmnElement).getDataState();
					if (dataState.length() == 0) {
						graph.getModel().setValue(cell, dataObject.getName());
					} else {
						graph.getModel().setValue(cell, dataObject.getName() + "\n[" + dataState + "]");
					}
				}
			}
		} else if (bpmnElement instanceof SequenceFlow) {
			String expression = ((SequenceFlow) bpmnElement).getConditionExpression().toValue();
			if (previous == null) {
				if (expression.length() == 0) {
					XMLComplexElement actOrGw = (XMLComplexElement) bpmnElementMap.get(cell.getSource().getId());
					if (graph.isDefaultSequenceFlow(cell)) {
						actOrGw.set("default", id);
					} else {
						actOrGw.set("default", "");
						graph.getModel().setStyle(cell, cell.getStyle() + ";startArrow=none;endArrow=classic;sourcePerimeterSpacing=2;expression= ");

					}
				} else {
					if (graph.isGateway(cell.getSource())) {
						graph.getModel().setStyle(cell,
								cell.getStyle() + ";startArrow=none;endArrow=classic;sourcePerimeterSpacing=2" + ";expression=" + expression);
					} else {
						graph.getModel().setStyle(cell,
								cell.getStyle() + ";startArrow=open_diamond;endArrow=classic;sourcePerimeterSpacing=15" + ";expression=" + expression);
					}
				}
			} else {
				XMLComplexElement actOrGw = (XMLComplexElement) bpmnElementMap.get(cell.getSource().getId());
				if (graph.isDefaultSequenceFlow(cell)) {
					actOrGw.set("default", id);
				} else {
					actOrGw.set("default", "");
					((SequenceFlow) bpmnElement).getConditionExpression().setValue(
							graph.getCellStyle(cell).get("expression") == null ? "" : (String) graph.getCellStyle(cell).get("expression"));
				}
			}
		} else if (bpmnElement instanceof Participant) {
			if (graph.isMultiInstanceParticipant(cell)) {
				((Participant) bpmnElement).setMultiplicity("2");
			} else {
				((Participant) bpmnElement).setMultiplicity("1");
			}

			if (graph.isChoreographyParticipant(cell)) {
				double height = 0;
				Object subprocess = null;
				Object parent = cell.getParent();
				Object[] partCells = mxGraphModel.getChildVertices(model, parent);
				for (int i = 0; i < partCells.length; i++) {
					if (graph.isChoreographyParticipant(partCells[i])) {
						mxCell part = (mxCell) partCells[i];
						mxGeometry geo = model.getGeometry(part);
						if (graph.isMultiInstanceParticipant(part)) {
							geo.setHeight(Constants.PARTICIPANT_HEIGHT * 1.75);
						} else {
							geo.setHeight(Constants.PARTICIPANT_HEIGHT);
						}
						model.setGeometry(part, geo);
						height += geo.getHeight();
					} else {
						subprocess = partCells[i];
					}
				}

				mxGeometry geo = model.getGeometry(parent);
				geo.setHeight(model.getGeometry(subprocess).getHeight() + height);
				model.setGeometry(parent, geo);

				Utils.arrangeChoreography(graph, parent, false);
				graph.refresh();
			}

		}
	}

	private void changeCellLabel(mxCell cell, String previous) {
		String label = cell.getValue().toString();
		String id = cell.getId();
		if (graph.isChoreographyParticipant(cell)) {
			id = id.substring(id.indexOf("_part_") + 6);
		}
		XMLComplexElement bpmnElement = (XMLComplexElement) bpmnElementMap.get(id);

		if (bpmnElement instanceof TextAnnotation) {
			((TextAnnotation) bpmnElement).setText(label);
		} else if (bpmnElement instanceof Group) {
			((Group) bpmnElement).setCategoryValueRef(BPMNModelUtils.getCategoryValueId(graph.getBpmnModel(), label));
		} else if (bpmnElement instanceof DataStoreReference) {
			((DataStoreReference) bpmnElement).setDataStoreRef(BPMNModelUtils.getDataStoreId(graph.getBpmnModel(), label));
		} else if (bpmnElement instanceof DataObjectReference) {
			String dataObjectId = "";
			String dataState = "";
			if (label.length() != 0) {
				int index = label.lastIndexOf("\n[");
				if (index > 0) {
					dataState = label.substring(index + 2, label.length() - 1);
					label = label.substring(0, index);
				}
				dataObjectId = ((FlowElements) bpmnElement.getParent()).getAccessibleDataObjectId(label, graph.isCollectionDataObject(cell));
			}
			((DataObjectReference) bpmnElement).setDataObjectRef(dataObjectId);
			if (previous != null) {
				((DataObjectReference) bpmnElement).setDataState(dataState);
			}
		} else if (bpmnElement instanceof Event) {
			XMLElement el = ((Event) bpmnElement).getEventDefinition();
			if (el instanceof LinkEventDefinition) {
				((LinkEventDefinition) el).setName(label);
			}
			bpmnElement.set("name", label);
		} else if (bpmnElement instanceof DataInput || bpmnElement instanceof DataOutput) {
			if (label != null && label.length() != 0) {
				int index = label.lastIndexOf("\n[");
				if (index > 0) {
					String name = label.substring(0, index);
					String state = label.substring(index + 2, label.length() - 1);
					bpmnElement.set("name", name);
					((ItemAwareElement) bpmnElement).setDataState(state);
				} else {
					bpmnElement.set("name", label);
					((ItemAwareElement) bpmnElement).setDataState("");
				}
			}
		} else if (bpmnElement != null) {
			bpmnElement.set("name", label);
		}
	}

	private void changeCell(mxCell cell) {
		String id = cell.getId();

		if (graph.isChoreography(cell)) {
			id += "_CT";
		} else if (graph.isSubChoreography(cell)) {
			id += "_SC";
		} else if (graph.isChoreographyParticipant(cell)) {
			id = id.substring(id.indexOf("_part_") + 6);
		}
		XMLElement bpmnElement = bpmnElementMap.get(id);
		if (bpmnElement == null) {
			addElement(cell);
		} else {
			if (graph.isChoreographyTask(cell) || graph.isChoreographySubprocess(cell)) {
				changeCellLoopType(cell);
			} else {
				Documentations documentations = null;
				XMLTextElements categoryValueRefs = null;
				if (bpmnElement instanceof BaseElement) {
					documentations = ((BaseElement) bpmnElement).getDocumentations();
					if (bpmnElement instanceof FlowElement) {
						categoryValueRefs = ((FlowElement) bpmnElement).getCategoryValueRefs();
					}
				}

				EventDefinitions eventDefinitions = null;
				XMLTextElements eventDefinitionRefs = null;
				Properties properties = null;
				ResourceRoles resourceRoles = null;
				LoopCharacteristics loopCharacteristics = null;
				if (bpmnElement instanceof Activity) {
					properties = ((Activity) bpmnElement).getProperties();
					loopCharacteristics = ((Activity) bpmnElement).getLoopCharacteristics();
					resourceRoles = ((Activity) bpmnElement).getResourceRoles();
				} else if (bpmnElement instanceof Event) {
					eventDefinitions = ((Event) bpmnElement).getEventDefinitions();
					eventDefinitionRefs = ((Event) bpmnElement).getEventDefinitionRefs();
					properties = ((Event) bpmnElement).getProperties();
				} else if (bpmnElement instanceof BPMNProcess) {
					properties = ((BPMNProcess) bpmnElement).getProperties();
				}

				removeCell(cell);
				removeEmptyProcess(bpmnElement);
				addElement(cell);

				bpmnElement = bpmnElementMap.get(id);
				if (graph.isChoreographyParticipant(cell)) {
					changeCellAttr(cell, null);
				}
				if (bpmnElement instanceof BaseElement) {
					((BaseElement) bpmnElement).setDocumentations(documentations);
					if (bpmnElement instanceof FlowElement) {
						((FlowElement) bpmnElement).setCategoryValueRefs(categoryValueRefs);
					}
				}
				if (bpmnElement instanceof Activity) {
					String loopType = "";
					if (graph.isMultiInstanceParallelActivity(cell)) {
						loopType = LoopTypes.TYPE_MULTI_INSTANCE_LOOPCHARACTERISTICS;
					} else if (graph.isMultiInstanceSequentialActivity(cell)) {
						loopType = LoopTypes.TYPE_MULTI_INSTANCE_LOOPCHARACTERISTICS;
					} else if (graph.isStandardLoopActivity(cell)) {
						loopType = LoopTypes.TYPE_STANDARD_LOOPCHARACTERISTICS;
					}
					if (loopCharacteristics == null && loopType.length() != 0 || loopCharacteristics != null && !loopCharacteristics.toName().equals(loopType)) {
						loopCharacteristics = ((Activity) bpmnElement).setLoopCharacteristics(loopType);
						if (loopCharacteristics instanceof MultiInstanceLoopCharacteristics) {
							if (graph.isMultiInstanceSequentialActivity(cell)) {
								((MultiInstanceLoopCharacteristics) loopCharacteristics).setIsSequential(true);
							}
						}
					} else {
						((Activity) bpmnElement).setLoopCharacteristics(loopCharacteristics);
					}
					((Activity) bpmnElement).setProperties(properties);
					((Activity) bpmnElement).setResourceRoles(resourceRoles);
				} else if (bpmnElement instanceof Event) {
					if (eventDefinitions != null) {
						if (((Event) bpmnElement).getEventDefinitions().isEmpty() && !graph.isNoneEvent(cell)) {
							((Event) bpmnElement).setEventDefinitions(eventDefinitions);
							((Event) bpmnElement).setEventDefinitionRefs(eventDefinitionRefs);
						} else if (eventDefinitions.size() == 1 && !(eventDefinitions.getXMLElements().get(0) instanceof LinkEventDefinition)
								&& ((Event) bpmnElement).getEventDefinition() != null
								&& eventDefinitions.getXMLElements().get(0).toName().equals(((Event) bpmnElement).getEventDefinition().toName())) {
							((Event) bpmnElement).setEventDefinitions(eventDefinitions);
						}
					}
					((Event) bpmnElement).setProperties(properties);
				} else if (bpmnElement instanceof BPMNProcess) {
					((BPMNProcess) bpmnElement).setProperties(properties);
				}
			}
		}
	}

	private void removeFromPreviousParent(mxCell previous, mxCell cell) {
		if (graph.isLane(previous)) {
			Lane lane = (Lane) bpmnElementMap.get(previous.getId());
			if (lane != null) {
				lane.removeFlowNodeRef(cell.getId());
			}
		}
	}
	
	private XMLElement removeCell(mxCell cell) {
		String id = cell.getId();
		if (id.endsWith("_act")) {
			id = id.substring(0, id.length() - 4);
		} else {
			if (graph.isChoreography(cell)) {
				id += "_CT";
			} else if (graph.isSubChoreography(cell)) {
				id += "_SC";
			} else if (graph.isChoreographyParticipant(cell)) {
				id = id.substring(id.indexOf("_part_") + 6);
			}
		}

		XMLElement bpmnElement = bpmnElementMap.remove(id);
		removeChildBpmnElements(cell);
		if (bpmnElement instanceof Participant) {
			if (graph.isChoreographyParticipant(cell)) {
				XMLElement ownerElement = null;
				mxCell owner = Utils.getChoreographyActivity(graph, cell.getParent());
				if (owner != null) {
					ownerElement = bpmnElementMap.get(owner.getId());
					((ChoreographyActivity) ownerElement).getParticipantRefs().remove(((Participant) bpmnElement).getId());
					if (graph.isAdditionalChoreographyParticipant(cell)) {
						Utils.arrangeChoreography(graph, cell.getParent(), false);
						graph.refresh();
					}
				}
				Set<ChoreographyActivity> acts = BPMNModelUtils.getChoreographyActivityByParticipantRef(bpmnModel, id);
				for (ChoreographyActivity act : acts) {
					if (act != ownerElement) {
						return null;
					}
				}
			} else {
				BPMNProcess process = (BPMNProcess) bpmnElementMap.remove(((Participant) bpmnElement).getProcessRef());
				if (process != null) {
					((XMLCollection) process.getParent()).remove(process.getId());
				}
			}
		} else if (bpmnElement instanceof ChoreographyActivity) {
			for (XMLElement participantRef : ((ChoreographyActivity) bpmnElement).getParticipantRefs().getXMLElements()) {
				if (!BPMNModelUtils.hasParticipantRef(bpmnElement, participantRef.toValue())) {
					bpmnElementMap.remove(participantRef.toValue());
					Choreography choreography = BPMNModelUtils.getChoreography(bpmnModel);
					if (choreography != null) {
						choreography.getParticipants().remove(participantRef.toValue());
					}
				}
			}
		} else if (bpmnElement instanceof Message) {
			if (graph.isAttachedMessage(cell)) {
				MessageFlow messageFlow = (MessageFlow) bpmnElementMap.get(cell.getParent().getId());
				if (messageFlow != null) {
					messageFlow.setMessageRef(null);
				}
			}
		} else if (bpmnElement instanceof MessageFlow) {

		} else if (bpmnElement instanceof Lane) {
			Object[] cells = graph.getChildCells(cell);
			if (cells != null && cells.length > 0) {
				for (Object child : cells) {
					removeCell((mxCell) child);
				}
			}
		} else if (bpmnElement instanceof SequenceFlow) {
			FlowNode source = (FlowNode) bpmnElementMap.get(((SequenceFlow) bpmnElement).getSourceRef());
			if (source != null) {
				source.removeOutgoing(((SequenceFlow) bpmnElement).getId());
			}
			FlowNode target = (FlowNode) bpmnElementMap.get(((SequenceFlow) bpmnElement).getTargetRef());
			if (target != null) {
				target.removeIncoming(((SequenceFlow) bpmnElement).getId());
			}
		} else if (bpmnElement instanceof FlowNode) {
			if (graph.isLane(cell.getParent())) {
				Lane lane = (Lane) bpmnElementMap.get(cell.getParent().getId());
				if (lane != null) {
					lane.removeFlowNodeRef(cell.getId());
				}
			}

			if (bpmnElement instanceof CallActivity) {
				// CallableElement process = (CallableElement) bpmnElementMap.remove(((CallActivity)
				// bpmnElement).getCalledElement());
				// ((XMLCollection) process.getParent()).remove(process.getId());
			}
		} else if (bpmnElement instanceof DataAssociation) {
			BaseElement source = (BaseElement) bpmnElementMap.get(((DataAssociation) bpmnElement).getSourceRef());
			BaseElement target = (BaseElement) bpmnElementMap.get(((DataAssociation) bpmnElement).getTargetRef());
			XMLComplexElement parentElement = (XMLComplexElement) bpmnElement.getParent().getParent();
			if (parentElement instanceof Activity) {
				InputOutputSpecification ios = ((Activity) parentElement).getIoSpecification();
				if (target instanceof DataInput) {
					ios.removeDataInputRefs(target.getId());
				}
				if (source instanceof DataOutput) {
					ios.removeDataOutputRefs(source.getId());
				}
			} else if (parentElement instanceof CatchEvent) {
				((CatchEvent) parentElement).removeDataOutputRefs(source.getId());
			} else if (parentElement instanceof ThrowEvent) {
				((ThrowEvent) parentElement).removeDataInputRefs(target.getId());
			}
			if (target instanceof DataInput) {
				((XMLCollection) target.getParent()).remove(target.getId());
			}
			if (source instanceof DataOutput) {
				((XMLCollection) source.getParent()).remove(source.getId());
			}
		} else if (bpmnElement instanceof DataInput) {
			InputOutputSpecification ios = (InputOutputSpecification) bpmnElement.getParent().getParent();
			ios.removeDataInputRefs(cell.getId());
		} else if (bpmnElement instanceof DataOutput) {
			InputOutputSpecification ios = (InputOutputSpecification) bpmnElement.getParent().getParent();
			ios.removeDataOutputRefs(cell.getId());
		} else if (bpmnElement instanceof Association) {
			XMLElement parent = bpmnElement.getParent().getParent();
			if (parent instanceof Collaboration) {
				if (bpmnElementMap.get(((Collaboration) parent).getId()) != parent) {
					BPMNModelUtils.getDefaultProcess(bpmnModel).getArtifacts().remove(((Association) bpmnElement).getId());
				}
			}
		}

		if (bpmnElement != null) {
			if (model.isEdge(cell)) {
				model.getCells().remove(cell.getId());
			}

			XMLElement parent = bpmnElement.getParent();
			if (parent instanceof XMLComplexElement) {
				((XMLComplexElement) parent).remove(id);
			} else {
				((XMLCollection) parent).remove(id);
			}

			if (bpmnElement instanceof ChoreographyTask) {
				Choreography choreography = BPMNModelUtils.getChoreography(bpmnModel);
				for (XMLElement mf : choreography.getMessageFlowList()) {
					if (!BPMNModelUtils.hasMessageFlowRef(choreography, ((MessageFlow) mf).getId())) {
						choreography.getMessageFlows().remove(((MessageFlow) mf).getId());
					}
				}
			}

			parent = parent.getParent();
			removeEmptyCollaboration(parent);
			removeEmptyChoreography(bpmnElement);
			removeEmptyProcess(bpmnElement);
		}

		Object[] attechedEvents = graph.getAttechedEvents(cell);
		if (attechedEvents != null && attechedEvents.length > 0) {
			for (Object event : attechedEvents) {
				removeCell((mxCell) event);
			}
		}
		Object message = graph.getAttechedMessage(cell);
		if (message != null) {
			removeCell((mxCell) message);
		}

		return bpmnElement;
	}

	private void removeChildBpmnElements(Object cell) {
		for (Object child : YGraphModel.getChildren(model, cell)) {
			removeChildBpmnElements(child);
			bpmnElementMap.remove(((mxCell) child).getId());
		}
	}

	private void removeEmptyCollaboration(XMLElement bpmnElement) {
		if (bpmnElement instanceof Collaboration && bpmnElement.isEmpty()) {
			bpmnModel.removeRootElement(((Collaboration) bpmnElement).getId());
			bpmnElementMap.remove(((Collaboration) bpmnElement).getId());

			BaseElement rootElement = (BaseElement) bpmnModel.getRootElement("C_1");
			if (rootElement != null) {
				bpmnModel.removeRootElement("C_1");
				rootElement.setId("_1");
			} else {
				rootElement = (BaseElement) bpmnModel.getRootElement("P_1");
				if (rootElement != null) {
					bpmnModel.removeRootElement("P_1");
					rootElement.setId("_1");
				}
			}
			if (rootElement != null) {
				bpmnModel.addRootElement(rootElement);
				bpmnElementMap.put(rootElement.getId(), rootElement);
			}
		}
	}

	private void removeEmptyChoreography(XMLElement bpmnElement) {
		if (bpmnElement instanceof ChoreographyActivity && !BPMNModelUtils.hasChoreographyActivity(bpmnModel)) {
			Choreography choreography = BPMNModelUtils.getParentChoreography(bpmnElement);
			bpmnElementMap.remove(choreography.getId());
			bpmnModel.removeRootElement(choreography.getId());

			BPMNProcess process = bpmnModel.createProcess(BPMNModelUtils.getCollaboration(bpmnModel) == null);
			process.addFlowElements(choreography.getFlowElements());
			process.addArtifacts(choreography.getArtifacts());
			if (process.isEmptyProcess()) {
				((XMLCollection) process.getParent()).remove(process.getId());
			} else {
				bpmnElementMap.put(process.getId(), process);
			}
		}
	}

	private void removeEmptyProcess(XMLElement bpmnElement) {
		BPMNProcess process = BPMNModelUtils.getParentProcess(bpmnElement);
		if (process != null && process.isEmptyProcess() && !BPMNModelUtils.hasProcessRef(BPMNModelUtils.getDefinitions(process), process.getId())) {
			((XMLCollection) process.getParent()).remove(process.getId());
			bpmnElementMap.remove(process.getId());

			String participantId = BPMNModelUtils.getParticipantByProcessId(process.getId(), BPMNModelUtils.getDefinitions(process));
			XMLElement participant = bpmnElementMap.get(participantId);
			if (participant != null && participant instanceof Participant) {
				((Participant) participant).setProcessRef(null);
			}
		}
	}

	private void resetLinkEvent(mxCell cell) {
		Event event = (Event) bpmnElementMap.get(cell.getId());
		if (event == null) {
			return;
		}
		LinkEventDefinition ed = (LinkEventDefinition) event.getEventDefinition();
		List<XMLElement> links = event.getParent().getLinkEventDefinition(event);
		if (!links.isEmpty()) {
			if (event instanceof ThrowEvent) {
				LinkEventDefinition targetLink = (LinkEventDefinition) links.get(0);
				targetLink.getSources().clear();
				ed.setTarget(targetLink.getId());
				Event targetEvent = (Event) targetLink.getParent().getParent();
				links = targetEvent.getParent().getLinkEventDefinition(targetEvent);
				targetLink.getSources().addValues(links);
			} else {
				ed.getSources().clear();
				ed.getSources().addValues(links);
				for (XMLElement sourceLink : links) {
					((LinkEventDefinition) sourceLink).setTarget(ed.getId());
				}
			}
		} else {
			if (event instanceof ThrowEvent) {
				ed.setTarget("");
			} else {
				ed.getSources().clear();
			}
		}
	}

	private FlowElementsContainer getFlowElementsContainer(mxCell cell) {
		FlowElementsContainer container = null;
		mxCell parent = graph.getParentCell(cell);
		if (graph.isPlane(parent)) {
			if (graph.isChoreographyTask(cell) || graph.isChoreographySubprocess(cell) || graph.isChoreographyParticipant(cell)) {
				container = BPMNModelUtils.getChoreography(bpmnModel);
				if (container == null) {
					container = bpmnModel.createChoreography();
					bpmnElementMap.put(((Choreography) container).getId(), (XMLElement) container);
				}
			} else if (graph.isSwimlane(cell)) {
				return container;
			} else {
				container = BPMNModelUtils.getChoreography(bpmnModel);
				if (container == null) {
					if (graph.isCallActivityProcess(parent)) {
						container = (FlowElementsContainer) bpmnElementMap.get(((CallActivity) bpmnElementMap.get(parent.getId())).getCalledElement());
					} else {
						container = BPMNModelUtils.getDefaultProcess(bpmnModel);
					}
					if (container == null) {
						container = bpmnModel.createProcess(BPMNModelUtils.getCollaboration(bpmnModel) == null);
						bpmnElementMap.put(((BPMNProcess) container).getId(), (XMLElement) container);
					}
				}
			}
		} else if (graph.isSubProcess(parent)) {
			container = (FlowElementsContainer) bpmnElementMap.get(parent.getId());
		} else if (graph.isSwimlane(parent)) {
			if (graph.isPool(parent)) {
				container = getProcessByParticipant(parent.getId());
			} else if (graph.isLane(parent)) {
				Lane lane = (Lane) bpmnElementMap.get(parent.getId());
				if (graph.isFlowNodeRef(cell)) {
					lane.addFlowNodeRef(cell.getId());
				}
				container = BPMNModelUtils.getParentProcess(lane);
			}
		}
		return container;
	}

	private BPMNProcess getProcessByParticipant(String participantId) {
		BPMNProcess process = null;
		Participant participant = (Participant) bpmnElementMap.get(participantId);
		if (participant != null) {
			String processId = participant.getProcessRef();
			process = (BPMNProcess) bpmnElementMap.get(processId);
		}
		if (process == null) {
			process = (BPMNProcess) bpmnModel.generateRootElement(RootElements.TYPE_PROCESS);
			process.setId("P" + participant.getId());
			//process.setProcessType(BPMNProcess.ProcessType.Public.toString());
			process.setIsExecutable(true);
			bpmnModel.addRootElement(process);
			bpmnElementMap.put(process.getId(), process);
			participant.setProcessRef(process.getId());
		}
		return process;
	}

	private String getConversationNodeType(Object cell) {
		String type = "";
		if (graph.isCallConversation(cell)) {
			type = ConversationNodes.TYPE_CALL_CONVERSATION;
		} else if (graph.isSubConversation(cell)) {
			type = ConversationNodes.TYPE_SUBCONVERSATION;
		} else {
			type = ConversationNodes.TYPE_CONVERSATION;
		}
		return type;
	}

	private String getFlowElementType(Object cell) {
		String type = "";
		if (graph.isStartEvent(cell)) {
			type = FlowElements.TYPE_START_EVENT;
		} else if (graph.isAttachedIntermediateEvent(cell)) {
			type = FlowElements.TYPE_BOUNDARY_EVENT;
		} else if (graph.isIntermediateEvent(cell)) {
			if (graph.isThrowEvent(cell)) {
				type = FlowElements.TYPE_INTERMEDIATE_THROW_EVENT;
			} else {
				type = FlowElements.TYPE_INTERMEDIATE_CATCH_EVENT;
			}
		} else if (graph.isEndEvent(cell)) {
			type = FlowElements.TYPE_END_EVENT;
		} else if (graph.isCallChoreographyActivity(cell)) {
			type = FlowElements.TYPE_CALL_CHOREOGRAPHY;
		} else if (graph.isChoreographyTask(cell)) {
			type = FlowElements.TYPE_CHOREOGRAPHY_TASK;
		} else if (graph.isChoreographySubprocess(cell)) {
			type = FlowElements.TYPE_SUBCHOREOGRAPHY;
		} else if (graph.isCallActivity(cell) || graph.isCallActivityProcess(cell)) {
			type = FlowElements.TYPE_CALL_ACTIVITY;
		} else if (graph.isAbstractTask(cell)) {
			type = FlowElements.TYPE_TASK;
		} else if (graph.isSendTask(cell)) {
			type = FlowElements.TYPE_SEND_TASK;
		} else if (graph.isReceiveTask(cell)) {
			type = FlowElements.TYPE_RECEIVE_TASK;
		} else if (graph.isServiceTask(cell)) {
			type = FlowElements.TYPE_SERVICE_TASK;
		} else if (graph.isUserTask(cell)) {
			type = FlowElements.TYPE_USER_TASK;
		} else if (graph.isScriptTask(cell)) {
			type = FlowElements.TYPE_SCRIPT_TASK;
		} else if (graph.isManualTask(cell)) {
			type = FlowElements.TYPE_MANUAL_TASK;
		} else if (graph.isBusinessRuleTask(cell)) {
			type = FlowElements.TYPE_BUSINESS_RULE_TASK;
		} else if (graph.isComplexGateway(cell)) {
			type = FlowElements.TYPE_COMPLEX_GATEWAY;
		} else if (graph.isGatewayWithoutMarker(cell) || graph.isExclusiveGateway(cell)) {
			type = FlowElements.TYPE_EXCLUSIVE_GATEWAY;
		} else if (graph.isInclusiveGateway(cell)) {
			type = FlowElements.TYPE_INCLUSIVE_GATEWAY;
		} else if (graph.isParallelGateway(cell)) {
			type = FlowElements.TYPE_PARALLEL_GATEWAY;
		} else if (graph.isEventGateway(cell)) {
			type = FlowElements.TYPE_EVENT_BASED_GATEWAY;
		} else if (graph.isTransactionSubProcess(cell)) {
			type = FlowElements.TYPE_TRANSACTION;
		} else if (graph.isAdhocSubProcess(cell)) {
			type = FlowElements.TYPE_ADHOC_SUBPROCESS;
		} else if (graph.isSubProcess(cell) || graph.isEventSubProcess(cell)) {
			type = FlowElements.TYPE_SUBPROCESS;
		} else if (graph.isDataObject(cell)) {
			type = FlowElements.TYPE_DATAOBJECT_REFERENCE;
		} else if (graph.isDataStore(cell)) {
			type = FlowElements.TYPE_DATASTORE_REFERENCE;
		}

		return type;
	}

	private String getEventDefinitionType(Object cell) {
		String type = "";
		if (graph.isCancelEvent(cell)) {
			type = EventDefinitions.TYPE_CANCEL_EVENT_DEFINITION;
		} else if (graph.isCompensationEvent(cell)) {
			type = EventDefinitions.TYPE_COMPENSATE_EVENT_DEFINITION;
		} else if (graph.isConditionalEvent(cell)) {
			type = EventDefinitions.TYPE_CONDITIONAL_EVENT_DEFINITION;
		} else if (graph.isErrorEvent(cell)) {
			type = EventDefinitions.TYPE_ERROR_EVENT_DEFINITION;
		} else if (graph.isEscalationEvent(cell)) {
			type = EventDefinitions.TYPE_ESCALATION_EVENT_DEFINITION;
		} else if (graph.isLinkEvent(cell)) {
			type = EventDefinitions.TYPE_LINK_EVENT_DEFINITION;
		} else if (graph.isMessageEvent(cell)) {
			type = EventDefinitions.TYPE_MESSAGE_EVENT_DEFINITION;
		} else if (graph.isSignalEvent(cell)) {
			type = EventDefinitions.TYPE_SIGNAL_EVENT_DEFINITION;
		} else if (graph.isTerminateEvent(cell)) {
			type = EventDefinitions.TYPE_TERMINATE_EVENT_DEFINITION;
		} else if (graph.isTimerEvent(cell)) {
			type = EventDefinitions.TYPE_TIMER_EVENT_DEFINITION;
		}
		return type;
	}

	public final mxIEventListener getHandler() {
		return handler;
	}

	public final Map<String, XMLElement> getBpmnElementMap() {
		return bpmnElementMap;
	}

	public final void setBpmnElementMap(Map<String, XMLElement> bpmnElementMap) {
		this.bpmnElementMap = bpmnElementMap;
	}

	public final Definitions getBpmnModel() {
		return bpmnModel;
	}

	public final void setBpmnModel(Definitions bpmnModel) {
		this.bpmnModel = bpmnModel;
	}

}
