package org.yaoqiang.bpmn.model.elements;

import java.util.ArrayList;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * XMLComplexElement
 * 
 * @author Shi Yaoqiang(shi_yaoqiang@yahoo.com)
 */
public abstract class XMLComplexElement extends XMLElement {

	protected LinkedHashMap<String, XMLElement> elements = new LinkedHashMap<String, XMLElement>();

	public XMLComplexElement(XMLElement parent) {
		super(parent);
		fillStructure();
	}

	public XMLComplexElement(XMLElement parent, String name) {
		super(parent, name);
		fillStructure();
	}

	protected abstract void fillStructure();

	public void set(String name, String value) {
		XMLElement el = get(name);
		if (el != null) {
			el.setValue(value);
		} else {
			throw new RuntimeException("No such element!");
		}
	}

	public void add(XMLElement el) {
		elements.put(el.toName(), el);
	}

	public XMLElement remove(String name) {
		return elements.remove(name);
	}

	public boolean isEmpty() {
		boolean isEmpty = true;
		for (XMLElement el : elements.values()) {
			isEmpty = isEmpty && el.isEmpty();
		}
		isEmpty = isEmpty && value.trim().length() == 0;
		return isEmpty;
	}

	public List<XMLElement> toElements() {
		return new ArrayList<XMLElement>(elements.values());
	}

	public List<XMLElement> getXMLElements() {
		List<XMLElement> els = new ArrayList<XMLElement>();
		for (XMLElement el : elements.values()) {
			if (!(el instanceof XMLAttribute)) {
				els.add(el);
			}
		}
		return els;
	}

	public XMLElement get(String name) {
		return elements.get(name);
	}

	public boolean containsName(String name) {
		return elements.containsKey(name);
	}

	public Object clone() {
		XMLComplexElement d = (XMLComplexElement) super.clone();
		d.elements = new LinkedHashMap<String, XMLElement>();
		for (Map.Entry<String, XMLElement> entry : elements.entrySet()) {
			String key = new String(entry.getKey());
			XMLElement value = (XMLElement) entry.getValue().clone();
			value.setParent(d);
			d.elements.put(key, value);
		}
		return d;
	}

}
