package org.yaoqiang.bpmn.model.elements.events;

import java.util.List;

import org.yaoqiang.bpmn.model.elements.XMLAttribute;
import org.yaoqiang.bpmn.model.elements.XMLElement;
import org.yaoqiang.bpmn.model.elements.XMLTextElement;
import org.yaoqiang.bpmn.model.elements.XMLTextElements;
import org.yaoqiang.bpmn.model.elements.core.common.FlowElements;
import org.yaoqiang.bpmn.model.elements.data.DataOutput;
import org.yaoqiang.bpmn.model.elements.data.DataOutputAssociation;
import org.yaoqiang.bpmn.model.elements.data.DataOutputAssociations;
import org.yaoqiang.bpmn.model.elements.data.DataOutputs;
import org.yaoqiang.bpmn.model.elements.data.OutputSet;

/**
 * CatchEvent
 * 
 * @author Shi Yaoqiang(shi_yaoqiang@yahoo.com)
 */
public abstract class CatchEvent extends Event {

	public CatchEvent(FlowElements parent, String name) {
		super(parent, name);
	}

	protected void fillStructure() {
		XMLAttribute attrParallelMultiple = new XMLAttribute(this, "parallelMultiple", Boolean.FALSE.toString());
		DataOutputs refDataOutputs = new DataOutputs(this);
		DataOutputAssociations refDataOutputAssociations = new DataOutputAssociations(this);
		OutputSet refOutputSet = new OutputSet(this);
		EventDefinitions refEventDefinitions = new EventDefinitions(this);
		XMLTextElements refEventDefinitionRef = new XMLTextElements(this, "eventDefinitionRef");

		super.fillStructure();
		add(attrParallelMultiple);
		add(refDataOutputs);
		add(refDataOutputAssociations);
		add(refOutputSet);
		add(refEventDefinitions);
		add(refEventDefinitionRef);
	}

	public final boolean isParallelMultiple() {
		return Boolean.parseBoolean(get("parallelMultiple").toValue());
	}

	public final DataOutputs getDataOutputs() {
		return (DataOutputs) get("dataOutputs");
	}

	public final List<XMLElement> getDataOutputList() {
		return getDataOutputs().getXMLElements();
	}

	public final DataOutput getDataOutput(String id) {
		return (DataOutput) getDataOutputs().getCollectionElement(id);
	}

	public final DataOutputAssociations getDataOutputAssociations() {
		return (DataOutputAssociations) get("dataOutputAssociations");
	}

	public final List<XMLElement> getDataOutputAssociationList() {
		return getDataOutputAssociations().getXMLElements();
	}

	public final OutputSet getOutputSet() {
		return (OutputSet) get("outputSet");
	}

	public final void setParallelMultiple(boolean parallelMultiple) {
		set("parallelMultiple", String.valueOf(parallelMultiple));
	}

	public final DataOutput addDataOutput(String id) {
		DataOutput dataOutput = (DataOutput) getDataOutputs().generateNewElement();
		dataOutput.setId(id);
		getDataOutputs().add(dataOutput);
		addDataOutputRef(id);
		return dataOutput;
	}

	public final void addDataOutputRef(String id) {
		if (!getOutputSet().getDataOutputRefs().contains(id)) {
			XMLTextElement dataOutputRef = (XMLTextElement) getOutputSet().getDataOutputRefs().generateNewElement();
			dataOutputRef.setValue(id);
			getOutputSet().getDataOutputRefs().add(dataOutputRef);
		}
	}

	public final DataOutputAssociation addDataOutputAssociation(String id, String dataOutputId, String targetId) {
		DataOutputAssociation doa = (DataOutputAssociation) getDataOutputAssociations().generateNewElement();
		doa.setId(id);
		doa.addSourceRef(dataOutputId);
		doa.setTargetRef(targetId);
		getDataOutputAssociations().add(doa);
		return doa;
	}

	public final void removeDataOutputRefs(String value) {
		getOutputSet().getDataOutputRefs().remove(value);
	}

}
